# Copyright 2022 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""The following effective temperature estimators are provided:

- Maximum pseudo-likelihood is an efficient estimator for the temperature 
  describing a classical Boltzmann distribution P(x) = \exp(-H(x)/T)/Z(T) 
  given samples from that distribution, where H(x) is the classical energy 
  function. The following links describe features of the estimator in
  application to equilibrium distribution drawn from binary quadratic models and
  non-equilibrium distributions generated by annealing:
  https://www.jstor.org/stable/25464568
  https://doi.org/10.3389/fict.2016.00023 

- An effective temperature can be inferred assuming freeze-out during the 
  anneal at s=t/t_a, an annealing schedule, and a device physical temperature.
  Necessary device-specific properties are published for online solvers:
  https://docs.dwavesys.com/docs/latest/doc_physical_properties.html

""" 

import warnings
import numpy as np
import dimod
from scipy import optimize
from typing import Tuple

__all__ = ['effective_field', 'maximum_pseudolikelihood_temperature',
           'freezeout_effective_temperature', 'fast_effective_temperature']

def effective_field(bqm,
                    samples=None,
                    current_state_energy=False) -> (np.ndarray,list):
    '''Returns the effective field for all variables and all samples.
    
    The effective field with ``current_state_energy = False`` is the energy
    attributable to setting a variable to value 1, conditioned on fixed values 
    for all neighboring variables (relative to exclusion of the variable, and 
    associated energy terms, from the problem). 
    
    The effective field with ``current_state_energy = True`` is the energy gained
    by flipping the variable state against its current value (from say -1 to 1 
    in the Ising case, or 0 to 1 in the QUBO case). A positive value indicates 
    that the energy can be decreased by flipping the variable, hence the 
    variable is in a locally excited state.
    If all values are negative  (positive) within a sample, that sample is a 
    local minima (maxima).  

    Any BQM can be converted to an Ising model with
    
    .. math::
        H(s) = Constant + \sum_i h_i s_i + 0.5 \sum_{i,j} J_{i,j} s_i s_j

    with unique values of :math:`J` (symmetric) and :math:`h`. The sample dependent effect field on variable i, :math:`f_i(s)`, is then defined

    if current_state_energy == False:
        .. math::
            f_i(s) = h_i + \sum_j J_{i,j} s_j
    else:
        .. math:: 
            f_i(s) = 2 s_i [h_i + \sum_j J_{i,j}  s_j]
    
    Args:
        bqm (:obj:`dimod.BinaryQuadraticModel`): 
            Binary quadratic model.
        samples (samples_like or :obj:`.SampleSet`,optional):
            A collection of raw samples. `samples_like` is an extension of
            NumPy's array like structure. See :func:`dimod.sampleset.as_samples`.
            By default, a single sample with all +1 assignments is used. 
        current_state_energy (bool, optional, default=False): 
            By default, returns the effective field (the energy 
            contribution associated to a state assignment of 1). When set to 
            True, returns the energy lost in flipping the value of each 
            variable. Note current_state_energy is typically negative for 
            positive temperature samples, meaning energy is not decreased
            by flipping the spin against its current assignment.
    Returns:
        samples_like: 
            A Tuple of the effective_fields, and the variable labels.
            Effective fields are returned as a :obj:`numpy.ndarray`.
            Rows index samples, and columns index variables in the order
            returned by variable labels.

    Examples:
       For a ferromagnetic Ising chain :math:`H = - 0.5 \sum_i s_i s_{i+1}`
       and for a ground state sample (all +1), the energy lost when flipping
       any spin is equal to the number of couplers frustrated: -2 in the center
       of the chain (variables 1,2,..,N-2), and -1 at the end (variables 0 and N-1).
       
       >>> import dimod
       >>> import numpy as np
       >>> from dwave.system.temperatures import effective_field
       >>> N = 5
       >>> bqm = dimod.BinaryQuadraticModel.from_ising({}, {(i,i+1) : -0.5 for i in range(N-1)})
       >>> var_labels = list(range(N))
       >>> samples = (np.ones(shape=(1,N)), var_labels)
       >>> E = effective_field(bqm,samples,current_state_energy=True)
       >>> print('Cost to flip spin against current assignment', E)     
       Cost to flip spin against current assignment (array([[-1., -2., -2., -2., -1.]]), [0, 1, 2, 3, 4])
       
    '''
    if samples is None:
        samples = np.ones(shape=(1,bqm.num_variables))
        labels = bqm.variables
    else:
        samples, labels = dimod.sampleset.as_samples(samples)

    if bqm.vartype is dimod.BINARY:
        bqm = bqm.change_vartype('SPIN', inplace=False)
        samples = 2*samples - 1
    
    h, (irow, icol, qdata), offset = bqm.to_numpy_vectors(
        variable_order=labels)
    # eff_field = h + J*s OR diag(Q) + (Q-diag(Q))*b
    effective_fields = np.tile(h[np.newaxis,:],(samples.shape[0],1))
    for sI in range(samples.shape[0]):
        np.add.at(effective_fields[sI,:],irow,qdata*samples[sI,icol])
        np.add.at(effective_fields[sI,:],icol,qdata*samples[sI,irow])

    if current_state_energy is True:
        #Ising: eff_field = 2*s*(h + J*s)
        effective_fields = 2*samples*effective_fields
        
    return (effective_fields,labels)

def maximum_pseudolikelihood_temperature(bqm=None,
                                         sampleset=None,
                                         site_energy=None,
                                         num_bootstrap_samples=0,
                                         seed=None,
                                         T_guess=None,
                                         optimize_method='bisect',
                                         T_bracket=(1e-3,1000)) -> Tuple[float,np.ndarray]:
    '''Returns a sampling-based temperature estimate.
    
    The temperature T parameterizes the Boltzmann distribution as 
    :math:`P(x) = \exp(-H(x)/T)/Z(T)`, where :math:`P(x)` is a probability over a state space, 
    :math:`H(x)` is the energy function (BQM) and :math:`Z(T)` is a normalization. 
    Given a sample set (:math:`S`), a temperature estimate establishes the 
    temperature that is most likely to have produced the sample set.
    An effective temperature can be derived from a sample set by considering the
    rate of excitations only. A maximum-pseudo-likelihood (MPL) estimator 
    considers local excitations only, which are sufficient to establish a 
    temperature efficiently (in compute time and number of samples). If the BQM
    consists of uncoupled variables then the estimator is equivalent to a 
    maximum likelihood estimator.
    
    The effective MPL temperature is defined by the solution T to 
    
    .. math::
       0 = \sum_i \sum_{s \in S} f_i(s) \exp(f_i(s)/T), 
    
    where f is the energy lost in flipping spin i against its current 
    assignment (the effective field).
    
    The problem is a convex root solving problem, and is solved with SciPy 
    optimize.

    If the distribution is not Boltzmann with respect to the BQM provided, as
    may be the case for heuristic samplers (such as annealers), the temperature
    estimate can be interpreted as characterizing only a rate of local 
    excitations. In the case of sample sets obtained from D-Wave annealing 
    quantum computers the temperature can be identified with a physical 
    temperature via a late-anneal freeze-out phenomena.
    
    Args:
        bqm (:obj:`dimod.BinaryQuadraticModel`, optional):
            Binary quadratic model describing sample distribution.
            If ``bqm`` and ``site_energy`` are both None, then by default 
            100 samples are drawn using :class:`~dwave.system.samplers.DWaveSampler`, 
            with ``bqm`` defaulted as described.
        sampleset (:class:`dimod.SampleSet`, optional):
            A set of samples, assumed to be fairly sampled from
            a Boltzmann distribution characterized by ``bqm``.
        site_energy (samples_like, optional):
            A Tuple of effective fields and site labels.
            Derived from the ``bqm`` and ``sampleset`` if not provided.
        num_bootstrap_samples (int, optional, default=0):
            Number of bootstrap estimators to calculate.
        seed (int, optional)
            Seeds the bootstrap method (if provided) allowing reproducibility
            of the estimators.
        T_guess (float, optional):
            User approximation to the effective temperature, must be 
            a positive scalar value.
            Seeding the root-search method can enable faster convergence.
            By default, T_guess is ignored if it falls outside the range
            of ``T_bracket``.
        optimize_method (str,optional,default='bisect'):
            SciPy method used for optimization. Options are 'bisect' and
            None (the default SciPy optimize method). 
        T_bracket (list or Tuple of 2 floats, optional, default=(0.001,1000)):
            If excitations are absent, temperature is defined as zero, otherwise 
            this defines the range of Temperatures over which to attempt a fit when
            using the 'bisect' ``optimize_method`` (the default).

    Returns:
        Tuple of float and NumPy array:
            (T_estimate,T_bootstrap_estimates)

            *T_estimate*: a temperature estimate
            *T_bootstrap_estimates*: a numpy array of bootstrap estimators

    Examples:
       Draw samples from a D-Wave Quantum Computer for a large spin-glass 
       problem (random couplers J, zero external field h).
       Establish a temperature estimate by maximum pseudo-likelihood. 

       Note that due to the complicated freeze-out properties of hard models,
       such as large scale spin-glasses, deviation from a classical Boltzmann 
       distribution is anticipated.
       Nevertheless, the T estimate can always be interpreted as an estimator
       of local excitations rates. For example T will be 0 if only 
       local minima are returned (even if some of the local minima are not 
       ground states).
       
       >>> import dimod
       >>> from dwave.system.temperatures import maximum_pseudolikelihood_temperature
       >>> from dwave.system import DWaveSampler
       >>> from random import random
       >>> sampler = DWaveSampler() 
       >>> bqm = dimod.BinaryQuadraticModel.from_ising({}, {e : 1-2*random() for e in sampler.edgelist})
       >>> sampleset = sampler.sample(bqm, num_reads=100, auto_scale=False)
       >>> T,T_bootstrap =  maximum_pseudolikelihood_temperature(bqm,sampleset) 
       >>> print('Effective temperature ',T)    # doctest: +SKIP
       Effective temperature  0.24066488780293813
       
    See also:

        https://doi.org/10.3389/fict.2016.00023

        https://www.jstor.org/stable/25464568
        
    '''
    
    T_estimate = 0
    T_bootstrap_estimates = np.zeros(num_bootstrap_samples)

    #Check for largest local excitation in every sample, and over all samples
    if site_energy is None:
        if bqm is None or sampleset is None:
            raise ValueError('site_energy can only be derived if both'
                             'bqm and sampleset are provided as arguments')
        site_energy = effective_field(bqm,
                                      sampleset,
                                      current_state_energy = True)
    max_excitation = np.max(site_energy[0],axis=1)
    max_excitation_all =  np.max(max_excitation)
    if max_excitation_all <= 0:
        #There are no local excitations present in the sample set, therefore
        #the temperature is estimated as 0. 
        pass
    else:
        
        def d_mean_log_pseudo_likelihood(x):
            #Derivative of mean (w.r.t samples) log pseudo liklihood amounts
            #to local energy matching criteria
            #O = sum_i \sum_s f_i(s) P(s,i) #s = sample, i = variable index
            #f_i(s) is energy lost in flipping spin s_i against current assignment.
            #P(s,i) = 1/(1 + exp[x f_i(s)]), probability to flip against current state.
            #x = -1/T
            with warnings.catch_warnings():
                #Overflow errors are safe:
                warnings.simplefilter(action='ignore', category=RuntimeWarning)
                expFactor = np.exp(site_energy[0]*x)
            return np.sum(site_energy[0]/(1 + expFactor))
        
        #Ensures good gradient method, except pathological cases
        if T_guess is None:
            x0 = -1/max_excitation_all
        else:
            if T_guess < T_bracket[0]:
                x0 = -1/T_bracket[0]
            elif T_guess > T_bracket[1]:
                x0 = -1/T_bracket[1]
            else:
                x0 = -1/T_guess
        root_results = None
        if optimize_method == 'bisect':
            #Check T_bracket
            if not 0 <= T_bracket[0] < T_bracket[1]:
                raise ValueError('Bad T_bracket, must be positive ordered scalars.')
            #Convert to bisection bracket for -1/T
            bisect_bracket = [-1/T_bracket[i] for i in range(2)]
            
            if d_mean_log_pseudo_likelihood(bisect_bracket[0]) <0:
                warnings.warn(
                    'Temperature is less than T_bracket[0], or perhaps negative: ' 
                    'rescaling the Hamiltonian, modification of T_bracket[0], or '
                    'a change of the optimize_method (to None) '
                    'can resolve this issue assuming a numerical precision issue '
                    'induced by very large or small effective fields is not the cause. '
                    'Automated precision requirements mean that this routine works '
                    'best when excitations (effective fields) are O(1).')
                T_estimate = T_bracket[0]
            elif d_mean_log_pseudo_likelihood(bisect_bracket[1]) > 0:
                warnings.warn(
                    'Temperature is greater than T_bracket[1], or perhaps negative:'
                    'rescaling the Hamiltonian, modification of T_bracket[1] or '
                    'a change of the optimize_method (to None) '
                    'can resolve this issue assuming a numerical precision issue '
                    'induced by very large or small effective fields is not the cause. '
                    'Automated precision requirements mean that this routine works '
                    'best when excitations (effective fields) are O(1).')
                T_estimate = T_bracket[1]
            else:
                #Bounds are good:
                if x0 < bisect_bracket[0] or x0 > bisect_bracket[1]:
                    x0 = (bisect_bracket[0] + bisect_bracket[1])/2 
                root_results = optimize.root_scalar(f=d_mean_log_pseudo_likelihood, x0 = x0,
                                                    method=optimize_method, bracket=bisect_bracket)
                T_estimate = -1/(root_results.root)
        else:
            #For very large or small site energies this may be numerically
            #unstable, therefore bisection search is the default.
            def dd_mean_log_pseudo_likelihood(x):
                
                import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter(action='ignore', category=RuntimeWarning)
                    #overflow errors are harmless, +Inf is a safe value.
                    expFactor = np.exp(site_energy[0]*x)
                    #divide by zero (1/expFactor) and divide by +Inf errors are harmless
                    return np.sum(-site_energy[0]*site_energy[0]/(expFactor + 2 + 1/expFactor))
        
            root_results = optimize.root_scalar(f=d_mean_log_pseudo_likelihood, x0 = x0,
                                                fprime=dd_mean_log_pseudo_likelihood)
            T_estimate = -1/root_results.root
        if num_bootstrap_samples > 0:
            #By bootstrapping with respect to samples we 
            if root_results is not None:
                x0 = root_results.root
            prng = np.random.RandomState(seed)
            num_samples = site_energy[0].shape[0]
            for bs in range(num_bootstrap_samples):
                indices = np.random.choice(
                    num_samples,
                    num_bootstrap_samples,
                    replace=True)
                T_bootstrap_estimates[bs],_ = maximum_pseudolikelihood_temperature(
                    site_energy = (site_energy[0][indices,:],site_energy[1]),
                    num_bootstrap_samples = 0,
                    T_guess = T_estimate)
    
    return T_estimate, T_bootstrap_estimates

def freezeout_effective_temperature(freezeout_B, temperature, units_B = 'GHz', units_T = 'mK') -> float:
    '''Provides an effective temperature as a function of freezeout information.
    
    See https://docs.dwavesys.com/docs/latest/c_qpu_annealing.html for a 
    complete summary of D-Wave annealing quantum computer operation.
    

    A D-Wave annealing quantum computer is assumed to implement a Hamiltonian
    :math:`H(s) = B(s)/2 H_P - A(s)/2 H_D`, where: :math:`H_P` is the unitless 
    diagonal problem Hamiltonian,
    :math:`H_D` is the unitless driver Hamiltonian, :math:`B(s)` is the problem energy scale; A(s)
    is the driver energy scale, amd :math:`s` is the normalized anneal
    time :math:`s = t/t_a` (in [0,1]).
    Diagonal elements of :math:`H_P`, indexed by the spin state :math:`x`, are equal to
    the energy of a classical Ising spin system 
    
    .. math::
        E_{Ising}(x) = \sum_i h_i x_i + \sum_{i>j} J_{i,j} x_i x_j

    If annealing achieves a thermally equilibrated distribution 
    over decohered states at large :math:`s` where :math:`A(s) \ll B(s)`, 
    and dynamics stop abruptly at :math:`s=s^*`, the distribution of returned
    samples is well described by a Boltzmann distribution:

    .. math::
        P(x) = \exp(- B(s^*) R E_{Ising}(x) / 2 k_B T)
    
    where T is the physical temperature, and :math:`k_B` is the Boltzmann constant.
    R is a Hamiltonain rescaling factor, if a QPU is operated with auto_scale=False, 
    then R=1.
    The function calculates the unitless effective temperature as :math:`T_{eff} = 2 k_B T/B(s^*)`.

    Device temperature :math:`T`, annealing schedules {:math:`A(s)`, :math:`B(s)`} and 
    single-qubit freeze-out (:math:`s^*`, for simple uncoupled Hamltonians) are reported 
    device properties: https://docs.dwavesys.com/docs/latest/doc_physical_properties.html 
    These values (typically specified in mK and GHz) allows the calculation of an effective 
    temperature for simple Hamiltonians submitted to D-Wave quantum computers. Complicated 
    problems exploiting embeddings, or with many coupled variables, may freeze out at 
    different values of s or piecemeal). Large problems may have slow dynamics at small 
    values of s, so :math:`A(s)` cannot be ignored as a contributing factor to the distribution.

    Note that for QPU solvers this temperature estimate applies to problems
    submitted with no additional scaling factors (sampling with ``auto_scale = False``). 
    If ``auto_scale=True`` (default) additional scaling factors must be accounted for. 

    Args:
        freezeout_B (float):
             :math:`B(s^*)`, the problem Hamiltonian energy scale at freeze-out.

        temperature (float):
            :math:`T`, the physical temperature of the quantum computer.
        
        units_B (string, optional, 'GHz'):
            Units in which ``freezeout_B`` is specified. Allowed values:
            'GHz' (Giga-Hertz) and 'J' (Joules).

        units_T (string, optional, 'mK'):
            Units in which the ``temperature`` is specified. Allowed values:
            'mK' (milli-Kelvin) and 'K' (Kelvin).

    
    Returns:
        float : The effective (unitless) temperature. 
    
    Examples:

       This example uses the 
       published parameters <https://docs.dwavesys.com/docs/latest/doc_physical_properties.html>
       for the Advantage_system4.1 QPU solver as of November 22nd 2021: 
       :math:`B(s=0.612) = 3.91` GHz , :math:`T = 15.4`mK.
       
       >>> from dwave.system.temperatures import freezeout_effective_temperature
       >>> T = freezeout_effective_temperature(freezeout_B = 3.91, temperature = 15.4)
       >>> print('Effective temperature at single qubit freeze-out is', T)  # doctest: +ELLIPSIS
       Effective temperature at single qubit freeze-out is 0.164...
    
    See also:

        The function :class:`~dwave.system.temperatures.fast_effective_temperature` 
        estimates the temperature for single-qubit Hamiltonians, in approximate
        agreement with estimates by this function at reported single-qubit 
        freeze-out values :math:`s^*` and device physical parameters.
    '''
    
    #Convert units_B to Joules
    if units_B == 'GHz':
        h = 6.626068e-34 #J/Hz
        freezeout_B = freezeout_B *h
        freezeout_B *= 1e9
    elif units_B == 'J':
        pass
    else:
        raise ValueException("Units must be 'J' (Joules) "
                             "or 'mK' (milli-Kelvin)")
    
    if units_T == 'mK':
        temperature = temperature * 1e-3
    elif units_T == 'K':
        pass
    else:
        raise ValueException("Units must be 'K' (Kelvin) "
                             "or 'mK' (milli-Kelvin)")
    kB = 1.3806503e-23 # J/K
    
    return 2*temperature*kB/freezeout_B

def fast_effective_temperature(sampler=None, num_reads=None, seed=None,
                               h_range=(-1/6.1,1/6.1), sampler_params=None,
                               optimize_method=None,
                               num_bootstrap_samples=0) -> Tuple[np.float64,np.float64]:
    '''Provides an estimate to the effective temperature, :math:`T`, of a sampler.
    
    This function submits a set of single-qubit problems to a sampler and 
    uses the rate of excitations to infer a maximum-likelihood estimate of temperature.

    Args:
        sampler (:class:`dimod.Sampler`, optional, default=\ :class:`~dwave.system.samplers.DWaveSampler`):
            A dimod sampler. 

        num_reads (int, optional):
            Number of reads to use. Default is 100 if not specified in 
            ``sampler_params``.

        seed (int, optional):
            Seeds the problem generation process. Allowing reproducibility
            from pseudo-random samplers.

        h_range (float, optional, default = [-1/6.1,1/6.1]):
            Determines the range of external fields probed for temperature
            inference. Default is based on a D-Wave Advantage processor, where
            single-qubit freeze-out implies an effective temperature of 6.1 
            (see :class:`~dwave.system.temperatures.freezeout_effective_temperature`).
            The range should be chosen inversely proportional to the anticipated
            temperature for statistical efficiency, and to accomodate precision 
            and other nonidealities such as precision limitations.
            
        sampler_params (dict, optional):
            Any additional non-defaulted sampler parameterization. If 
            ``num_reads`` is a key, must be compatible with ``num_reads`` 
            argument.

        optimize_method (str, optional):
            Optimize method used by SciPy ``root_scalar`` method. The default
            method works well under default operation, 'bisect' can be 
            numerically more stable when operated without defaults.

        num_bootstrap_samples (int, optional, default=0):
            Number of bootstrap samples to use for estimation of the
            standard error. By default no bootstrapping is performed
            and the standard error is defaulted to 0.

    Returns:
        Tuple[float, float]:
            The effective temperature describing single qubit problems in an
            external field, and a standard error (+/- 1 sigma). 
            By default the confidence interval is set as 0.
             
    See also:

        https://doi.org/10.3389/fict.2016.00023

        https://www.jstor.org/stable/25464568
    
    Examples:
       Draw samples from a :class:`~dwave.system.samplers.DWaveSampler`, and establish the temperature
       
       >>> from dwave.system.temperatures import fast_effective_temperature
       >>> from dwave.system import DWaveSampler
       >>> sampler = DWaveSampler()
       >>> T, _ = fast_effective_temperature(sampler)
       >>> print('Effective temperature at freeze-out is',T)    # doctest: +SKIP
       0.21685104745347336
    
    See also:
        The function :class:`~dwave.system.temperatures.freezeout_effective_temperature` 
        may be used in combination with published device values to estimate single-qubit 
        freeze-out, in approximate agreement with empirical estimates of this function.

        https://doi.org/10.3389/fict.2016.00023

        https://www.jstor.org/stable/25464568
    
    '''
    
    if sampler is None:
        from dwave.system import DWaveSampler
        sampler = DWaveSampler()
        
    if 'h_range' in sampler.properties:
        if h_range[0] < sampler.properties['h_range'][0]:
            raise ValueError('h_range[0] exceeds programmable range')
           
        if h_range[1] > sampler.properties['h_range'][1]:
            raise ValueError('h_range[1] exceeds programmable range')
        
    prng = np.random.RandomState(seed)
    h_values = h_range[0] + (h_range[1]-h_range[0])*prng.rand(len(sampler.nodelist))
    bqm = dimod.BinaryQuadraticModel.from_ising({var: h_values[idx] for idx,var in enumerate(sampler.nodelist)}, {})
    
    #Create local sampling_params copy - default necessary additional fields:
    if sampler_params is None:
        sampler_params0 = {}
    else:
        sampler_params0 = sampler_params.copy()
    if num_reads is None:
        #Default is 100, makes efficient use of QPU access time:
        if 'num_reads' not in sampler_params0:
            sampler_params0['num_reads'] = 100 
    elif ('num_reads' in sampler_params0
        and sampler_params0['num_reads'] != num_reads):
        raise ValueError("sampler_params['num_reads'] != num_reads, "
                         "incompatible input arguments.")
    else:
        sampler_params0['num_reads'] = num_reads
    if ('auto_scale' in sampler_params0
        and sampler_params0['auto_scale'] is not False):
        raise ValueError("sampler_params['auto_scale'] == False, "
                         "is required by this method.")
    else:
        sampler_params0['auto_scale'] = False

    if num_bootstrap_samples is None:
        num_bootstrap_samples = sampler_params0['num_reads'] 
        
    sampleset = sampler.sample(bqm, **sampler_params0)

    T,Tboot = maximum_pseudolikelihood_temperature(
        bqm,
        sampleset,
        optimize_method=optimize_method,
        num_bootstrap_samples=num_bootstrap_samples)

    
    if num_bootstrap_samples == 0:
        return T, np.float64(0.0)
    else:
        return T, np.sqrt(np.var(Tboot))



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


def ramp(s, width, annealing_time):
    """Schedule with a ramp shape.

    Args:
        s (float):
            The mid-point of the ramp, as a fraction of the annealing time.

        width (float):
            The width of the ramp, as a fraction of the annealing time. Note
            that QPUs have a maximum slope.

        annealing_time (float):
            The total annealing time, in microseconds.

    Returns:
        list[2-tuple]: The points defining in a piece-wise curve in the shape of
        a ramp.

    Examples:
        This example constructs a schedule for a QPU that supports
        configuring an `h_gain_schedule`.

        >>> sampler = DWaveSampler(solver=dict(annealing_time=True,
                                               h_gain_schedule=True))
        >>> h = {v: -1 for v in sampler.nodelist}
        >>> schedule = ramp(.5, .2, sampler.properties['default_annealing_time'])
        >>> sampleset = sampler.sample_ising(h, {}, h_gain_schedule=schedule)

    """
    if s <= 0 or s >= 1:
        raise ValueError("s should be in interval (0, 1)")
    if width >= min(s, 1 - s) / 2:
        raise ValueError("given width takes curve outside of [0, 1] interval")

    return [(0, 0),
            (annealing_time * (s - width / 2), 0),
            (annealing_time * (s + width / 2), 1),
            (annealing_time, 1)]



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import numpy as np
import dimod

from dwave.system.temperatures import (maximum_pseudolikelihood_temperature,
                                       effective_field,
                                       freezeout_effective_temperature,
                                       fast_effective_temperature)

from dwave.system.testing import MockDWaveSampler

class TestTemperatures(unittest.TestCase):
    
    def test_effective_field(self):
        # For a simple model of independent spins H = sum_i s_i
        # The effective field is 1 (setting a spin to 1, costs 1 unit of energy,
        # relative to its exclusion)
        num_var = 3
        num_samples = 2
        var_labels = list(range(num_var))
        bqm = dimod.BinaryQuadraticModel.from_ising({var: 1 for var in var_labels}, {})
        samples_like = (np.ones(shape=(num_samples,num_var)),var_labels)
        E = effective_field(bqm,
                            samples_like)
        self.assertTrue(np.array_equal(np.ones(shape=(num_samples,num_var)), E[0]))
        self.assertTrue(num_var==len(E[1]))
        # energy lost in flipping from sample value (1) to -1 is H(1) - H(-1) = +2.
        E = effective_field(bqm,
                            samples_like,
                            current_state_energy=True)
        self.assertTrue(np.array_equal(2*np.ones(shape=(num_samples,num_var)), E[0]))

    def test_effective_field_vartype(self):
        # Check effective fields are identical whether using bqm or ising model
        num_var = 4
        var_labels = list(range(num_var))
        bqm = dimod.BinaryQuadraticModel.from_ising({var: var for var in var_labels}, {(var1,var2) : var1%2 + var2%2 - 1 for var1 in var_labels for var2 in var_labels})
        E_ising = effective_field(bqm,current_state_energy=True)
        bqm.change_vartype('BINARY',inplace=True)
        E_bqm = effective_field(bqm,current_state_energy=True)
        self.assertTrue(bqm.vartype==dimod.BINARY) 
        self.assertTrue(np.array_equal(E_ising[0], E_bqm[0]))

    def test_maximum_pseudolikelihood_temperature(self):
        # Single variable H = s_i problem with mean energy (-15 + 5)/20 = -0.5
        # 5 measured excitations out of 20.
        # This implies an effective temperature 1/atanh(0.5)
        site_energy = np.array([2]*5 + [-2]*15)
        site_names = ['a']
        for optimize_method in [None,'bisect']:
            T = maximum_pseudolikelihood_temperature(
                site_energy = (site_energy[:,np.newaxis],site_names),
                optimize_method = optimize_method)
            self.assertTrue(type(T) is tuple and len(T)==2)
            self.assertTrue(np.abs(T[0]-1/np.arctanh(0.5))<1e-8)
        
        # Single variable H = s_i problem with mean energy (-5 + 5)/10 = 0
        # This implies an infinite temperature (up to numerical tolerance
        # threshold of scipy optimize.)
        site_energy = np.array([1]*5 + [-1]*5)
        T_bracket = [0.1,1]
        with self.assertWarns(UserWarning) as w:
            # Returned value should match upper bracket value and
            # throw a warning.
            # Temperature is infinite (excitations and relaxations)
            # are equally likely.
            T = maximum_pseudolikelihood_temperature(
                site_energy = (site_energy[:,np.newaxis],[1]),
                T_bracket=T_bracket)
            self.assertTrue(type(T) is tuple and len(T)==2)
            self.assertTrue(T[0]==T_bracket[1])
        
        # Single variable H = s_i problem with no sample excitations 
        # This implies zero temperature 
        # Any bounds on T_bracket should be ignored. 
        site_energy = np.array([-1]*5)
        import warnings
        with warnings.catch_warnings():
            #Ignore expected 'out of T_bracket bound' warning:
            warnings.simplefilter(action='ignore', category=UserWarning)
            T = maximum_pseudolikelihood_temperature(
                site_energy = (site_energy[:,np.newaxis],[1]),
                T_bracket=T_bracket)
        self.assertTrue(type(T) is tuple and len(T)==2)
        self.assertTrue(T[0]==0)

    def test_freezeout_effective_temperature(self):
        # 24mK and 1GHz line up conveniently for T=1.00
        BsGHz=1
        TmK=24
        T = freezeout_effective_temperature(BsGHz,TmK)
        self.assertTrue(np.round(T*100)==100)
        
        # https://docs.dwavesys.com/docs/latest/doc_physical_properties.html
        # Accessed November 12th, 2021
        # Advantage_system4.1 (Advantage): B(s=0.612) = 3.91GHz , T = 15.4mK
        # T_eff = 0.16
        # DW_2000Q_6 (DW2000Q-LN): B(s=0.719) = 6.54GHz , T = 13.5mK
        # T_eff = 0.086 
        BsGHz=3.91
        TmK=15.4
        T = freezeout_effective_temperature(BsGHz,TmK)
        self.assertTrue(np.round(T*100)==16)
        
        BsGHz=6.54
        TmK=13.5
        T = freezeout_effective_temperature(BsGHz,TmK)
        self.assertTrue(np.round(T*1000)==86)
        
        
    def test_fast_effective_temperature(self):
        # Initializing in a ground state, all effective
        # fields must be non-negative.
        sampler = MockDWaveSampler()
        T, sigma = fast_effective_temperature(sampler=sampler)
        # Simulated Annealer will tend to return only ground
        # states, hence temperature 0. But exceptions are
        # possible. Hence no assertion on value.
        

    def test_bootstrap_errors(self):
        site_energy = np.array([2]*25 +  [-2]*75)
        num_bootstrap_samples = 100
        
        T,Tb = maximum_pseudolikelihood_temperature(site_energy = (site_energy[:,np.newaxis],[1]),num_bootstrap_samples = num_bootstrap_samples)
        
        # Add test to check bootstrap estimator implementation.
        # T = 1/np.arctanh(0.5). With high probability bootstrapped values
        # are finite and will throw no warnings.
        self.assertTrue(len(Tb) == num_bootstrap_samples)



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

__all__ = ['__version__', '__author__', '__authoremail__', '__description__']

__version__ = '1.16.0'
__author__ = 'D-Wave Systems Inc.'
__authoremail__ = 'tools@dwavesys.com'
__description__ = 'All things D-Wave System.'


# register the non-open-source packages required for dwave-drivers to work
contrib = [{
    'name': 'drivers',
    'title': 'D-Wave Drivers',
    'description': 'These drivers enable some automated performance-tuning features.',
    'license': {
        'name': 'D-Wave EULA',
        'url': 'https://docs.ocean.dwavesys.com/eula',
    },
    'requirements': [
        'dwave-drivers>=0.4.4',
    ]
}]



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
A :std:doc:`dimod composite <oceandocs:docs_dimod/reference/samplers>` that uses the D-Wave virtual
graph feature for improved :std:doc:`minor-embedding <oceandocs:docs_system/intro>`.

D-Wave *virtual graphs* simplify the process of minor-embedding by enabling you to more
easily create, optimize, use, and reuse an embedding for a given working graph. When you submit an
embedding and specify a chain strength using these tools, they automatically calibrate the qubits
in a chain to compensate for the effects of biases that may be introduced as a result of strong
couplings.

See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
for explanations of technical terms in descriptions of Ocean tools.
"""

import dimod

from dwave.system.composites.embedding import FixedEmbeddingComposite
from dwave.system.flux_bias_offsets import get_flux_biases


FLUX_BIAS_KWARG = 'flux_biases'

__all__ = ['VirtualGraphComposite']


class VirtualGraphComposite(FixedEmbeddingComposite):
    """Composite to use the D-Wave virtual graph feature for minor-embedding.

    Calibrates qubits in chains to compensate for the effects of biases and enables easy
    creation, optimization, use, and reuse of an embedding for a given working graph.

    Inherits from :class:`dimod.ComposedSampler` and :class:`dimod.Structured`.

    Args:
        sampler (:class:`.DWaveSampler`):
            A dimod :class:`dimod.Sampler`. Typically a :obj:`.DWaveSampler` or
            derived composite sampler; other samplers may not work or make sense with
            this composite layer.

        embedding (dict[hashable, iterable]):
            Mapping from a source graph to the specified sampler's graph (the target graph).

        chain_strength (float, optional, default=None):
            Desired chain coupling strength. This is the magnitude of couplings between qubits
            in a chain. If None, uses the maximum available as returned by a SAPI query
            to the D-Wave solver.

        flux_biases (list/False/None, optional, default=None):
            Per-qubit flux bias offsets in the form of a list of lists, where each sublist
            is of length 2 and specifies a variable and the flux bias offset associated with
            that variable. Qubits in a chain with strong negative J values experience a
            J-induced bias; this parameter compensates by recalibrating to remove that bias.
            If False, no flux bias is applied or calculated.
            If None, flux biases are pulled from the database or calculated empirically.

        flux_bias_num_reads (int, optional, default=1000):
            Number of samples to collect per flux bias value to calculate calibration
            information.

        flux_bias_max_age (int, optional, default=3600):
            Maximum age (in seconds) allowed for a previously calculated flux bias offset to
            be considered valid.

    .. attention::
       D-Wave's *virtual graphs* feature can require many seconds of D-Wave system time to calibrate
       qubits to compensate for the effects of biases. If your account has limited
       D-Wave system access, consider using :class:`.FixedEmbeddingComposite` instead.

    Examples:
       This example uses :class:`.VirtualGraphComposite` to instantiate a composed sampler
       that submits a QUBO problem to a D-Wave solver.
       The problem represents a logical
       AND gate using penalty function :math:`P = xy - 2(x+y)z +3z`, where variables x and y
       are the gate's inputs and z the output. This simple three-variable problem is manually
       minor-embedded to a single :std:doc:`Chimera <oceandocs:docs_system/intro>` unit cell:
       variables x and y are represented by qubits 1 and 5, respectively, and z by a
       two-qubit chain consisting of qubits 0 and 4.
       The chain strength is set to the maximum allowed found from querying the solver's extended
       J range. In this example, the ten returned samples all represent valid states of
       the AND gate.

       >>> from dwave.system import DWaveSampler, VirtualGraphComposite
       >>> embedding = {'x': {1}, 'y': {5}, 'z': {0, 4}}
       >>> qpu_2000q = DWaveSampler(solver={'topology__type': 'chimera'})
       >>> qpu_2000q.properties['extended_j_range']
       [-2.0, 1.0]
       >>> sampler = VirtualGraphComposite(qpu_2000q, embedding, chain_strength=2) # doctest: +SKIP
       >>> Q = {('x', 'y'): 1, ('x', 'z'): -2, ('y', 'z'): -2, ('z', 'z'): 3}
       >>> sampleset = sampler.sample_qubo(Q, num_reads=10) # doctest: +SKIP
       >>> print(sampleset)    # doctest: +SKIP
          x  y  z energy num_oc. chain_.
       0  1  0  0    0.0       2     0.0
       1  0  1  0    0.0       3     0.0
       2  1  1  1    0.0       3     0.0
       3  0  0  0    0.0       2     0.0
       ['BINARY', 4 rows, 10 samples, 3 variables]

    See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
    for explanations of technical terms in descriptions of Ocean tools.

    """

    def __init__(self, sampler, embedding,
                 chain_strength=None,
                 flux_biases=None,
                 flux_bias_num_reads=1000,
                 flux_bias_max_age=3600):

        super(VirtualGraphComposite, self).__init__(sampler, embedding)
        self.parameters.update(apply_flux_bias_offsets=[])

        # Validate the chain strength, or obtain it from J-range if chain strength is not provided.
        self.chain_strength = _validate_chain_strength(sampler, chain_strength)

        if flux_biases is False:  # use 'is' because bool(None) is False
            # in this case we are done
            self.flux_biases = None
            return

        if FLUX_BIAS_KWARG not in sampler.parameters:
            raise ValueError("Given child sampler does not accept flux_biases.")

        # come back as a dict
        flux_biases = get_flux_biases(sampler, embedding,
                                      num_reads=flux_bias_num_reads,
                                      chain_strength=self.chain_strength,
                                      max_age=flux_bias_max_age)

        self.flux_biases = [flux_biases.get(v, 0.0) for v in range(sampler.properties['num_qubits'])]

        return

    @dimod.bqm_structured
    def sample(self, bqm, apply_flux_bias_offsets=True, **kwargs):
        """Sample from the given Ising model.

        Args:

            h (list/dict):
                Linear biases of the Ising model. If a list, the list's indices
                are used as variable labels.

            J (dict of (int, int):float):
                Quadratic biases of the Ising model.

            apply_flux_bias_offsets (bool, optional):
                If True, use the calculated flux_bias offsets (if available).

            **kwargs:
                Optional keyword arguments for the sampling method, specified per solver.

        Examples:
           This example uses :class:`.VirtualGraphComposite` to instantiate a composed sampler
           that submits an Ising problem to a D-Wave solver.
           The problem represents a logical
           NOT gate using penalty function :math:`P = xy`, where variable x is the gate's input
           and y the output. This simple two-variable problem is manually minor-embedded
           to a single :std:doc:`Chimera <oceandocs:docs_system/intro>` unit cell: each variable
           is represented by a chain of half the cell's qubits, x as qubits 0, 1, 4, 5,
           and y as qubits 2, 3, 6, 7.
           The chain strength is set to half the maximum allowed found from querying the solver's extended
           J range. In this example, the ten returned samples all represent valid states of
           the NOT gate.

           >>> from dwave.system import DWaveSampler, VirtualGraphComposite
           >>> embedding = {'x': {0, 4, 1, 5}, 'y': {2, 6, 3, 7}}
           >>> qpu_2000q = DWaveSampler(solver={'topology__type': 'chimera'})
           >>> qpu_2000q.properties['extended_j_range']
           [-2.0, 1.0]
           >>> sampler = VirtualGraphComposite(qpu_2000q, embedding, chain_strength=1) # doctest: +SKIP
           >>> h = {}
           >>> J = {('x', 'y'): 1}
           >>> sampleset = sampler.sample_ising(h, J, num_reads=10) # doctest: +SKIP
           >>> print(sampleset)    # doctest: +SKIP
              x  y energy num_oc. chain_.
           0 -1 +1   -1.0       6     0.0
           1 +1 -1   -1.0       4     0.0
           ['SPIN', 2 rows, 10 samples, 2 variables]

        See `Ocean Glossary <https://docs.ocean.dwavesys.com/en/stable/concepts/index.html>`_
        for explanations of technical terms in descriptions of Ocean tools.

        """

        if apply_flux_bias_offsets:
            if self.flux_biases is not None:
                kwargs[FLUX_BIAS_KWARG] = self.flux_biases

        kwargs.setdefault('chain_strength', self.chain_strength)

        return super(VirtualGraphComposite, self).sample(bqm, **kwargs)


def _validate_chain_strength(sampler, chain_strength):
    """Validate the provided chain strength, checking J-ranges of the sampler's children.

    Args:
        chain_strength (float) The provided chain strength.  Use None to use J-range.

    Returns (float):
        A valid chain strength, either provided or based on available J-range.  Positive finite float.

    """
    properties = sampler.properties

    if 'extended_j_range' in properties:
        max_chain_strength = - min(properties['extended_j_range'])
    elif 'j_range' in properties:
        max_chain_strength = - min(properties['j_range'])
    else:
        raise ValueError("input sampler should have 'j_range' and/or 'extended_j_range' property.")

    if chain_strength is None:
        chain_strength = max_chain_strength
    elif chain_strength > max_chain_strength:
        raise ValueError("Provided chain strength exceedds the allowed range.")

    return chain_strength



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import os
import unittest

import dimod
import numpy as np

from dwave.cloud.exceptions import ConfigFileError, SolverNotFoundError
from dwave.system import LeapHybridDQMSampler

try:
    sampler = LeapHybridDQMSampler()
except (ValueError, ConfigFileError, SolverNotFoundError):
    sampler = None


@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
@unittest.skipIf(sampler is None, "no hybrid sampler available")
class TestLeapHybridSampler(unittest.TestCase):
    def test_smoke(self):
        dqm = dimod.DQM()
        u = dqm.add_variable(2, 'a')
        v = dqm.add_variable(3)
        dqm.set_linear(u, [1, 2])
        dqm.set_quadratic(u, v, {(0, 1): 1, (0, 2): 1})

        try:
            # dimod 0.10+
            dqm.offset += 3
        except AttributeError:
            pass

        sampleset = sampler.sample_dqm(dqm)

        np.testing.assert_array_almost_equal(dqm.energies(sampleset),
                                             sampleset.record.energy)

    def test_smoke_case_label(self):
        try:
            from dimod import CaseLabelDQM
        except ImportError:
            self.skipTest("need dimod 0.10+")

        dqm = CaseLabelDQM()
        u = dqm.add_variable({'red', 'green', 'blue'}, shared_labels=True)
        dqm.offset = 5

        sampleset = sampler.sample_dqm(dqm)

        np.testing.assert_array_almost_equal(dqm.energies(sampleset),
                                             sampleset.record.energy)

    # NOTE: enable when problem labelling deployed to prod
    @unittest.skipIf(sampler is not None and 'cloud' in sampler.client.endpoint,
                     "labels not supported in production")
    def test_problem_labelling(self):
        dqm = dimod.DQM()
        u = dqm.add_variable(2, 'a')
        v = dqm.add_variable(3)
        dqm.set_linear(u, [1, 2])
        dqm.set_quadratic(u, v, {(0, 1): 1, (0, 2): 1})
        label = 'problem label'

        sampleset = sampler.sample_dqm(dqm, label=label)
        self.assertIn('problem_id', sampleset.info)
        self.assertIn('problem_label', sampleset.info)
        self.assertEqual(sampleset.info['problem_label'], label)



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import itertools
import unittest
import unittest.mock

from parameterized import parameterized

import dimod
import dwave_networkx as dnx

from dwave.cloud import exceptions
from dwave.cloud import computation
from dwave.system import DWaveCliqueSampler, DWaveSampler
from dwave.system.exceptions import FailoverCondition, RetryCondition


class MockDWaveSampler(dimod.RandomSampler, dimod.Structured):
    # contains the minimum needed to work with DWaveCliqueSampler

    edgelist = None
    nodelist = None

    def __init__(self, **kwargs):
        self.properties = dict(h_range=[-2, 2],
                               j_range=[-1, 1],
                               extended_j_range=[-2, 1],)
        self.parameters = {'auto_scale': None}

    def sample(self, bqm, auto_scale=True):
        assert not auto_scale
        assert bqm.vartype is dimod.SPIN

        h_range = self.properties['h_range']
        j_range = self.properties['extended_j_range']

        for bias in bqm.linear.values():
            assert h_range[0] <= bias <= h_range[1]

        for bias in bqm.quadratic.values():
            assert j_range[0] <= bias <= j_range[1]

        return super().sample(bqm)

    def trigger_failover(self):
        pass

    to_networkx_graph = DWaveSampler.to_networkx_graph


class MockChimeraDWaveSampler(MockDWaveSampler):
    def __init__(self, **config):
        super().__init__()

        self.properties.update(topology=dict(shape=[4, 4, 4], type='chimera'))

        G = dnx.chimera_graph(4, 4, 4)

        self.nodelist = list(G.nodes)
        self.edgelist = list(G.edges)

    def sample(self, bqm, **kwargs):

        # per_qubit_coupling_range
        ran = (-9, 6)

        # check the total coupling range
        for v in bqm.variables:
            bias = sum(bqm.adj[v].values())
            assert ran[0] <= bias <= ran[1]

        return super().sample(bqm, **kwargs)


class MockPegasusDWaveSampler(MockDWaveSampler):
    def __init__(self, **config):
        super().__init__()

        self.properties.update(topology=dict(shape=[6], type='pegasus'))

        G = dnx.pegasus_graph(6)

        self.nodelist = list(G.nodes)
        self.edgelist = list(G.edges)


with unittest.mock.patch('dwave.system.samplers.clique.DWaveSampler',
                         MockChimeraDWaveSampler):
    chimera_sampler = DWaveCliqueSampler()

with unittest.mock.patch('dwave.system.samplers.clique.DWaveSampler',
                         MockPegasusDWaveSampler):
    pegasus_sampler = DWaveCliqueSampler()


@dimod.testing.load_sampler_bqm_tests(chimera_sampler)
@dimod.testing.load_sampler_bqm_tests(pegasus_sampler)
class TestDWaveCliqueSampler(unittest.TestCase):
    def test_api(self):
        dimod.testing.assert_sampler_api(chimera_sampler)
        dimod.testing.assert_sampler_api(pegasus_sampler)

    def test_clique(self):
        self.assertEqual(len(chimera_sampler.clique(2)), 2)
        self.assertEqual(len(chimera_sampler.clique(16)), 16)

        self.assertEqual(len(chimera_sampler.clique(['a', 1])), 2)
        self.assertEqual(set(chimera_sampler.clique(['a', 1])), {'a', 1})

    def test_largest_clique(self):
        self.assertEqual(len(chimera_sampler.largest_clique()), 16)

    def test_ferromagnet_chimera(self):
        # submit a maximum ferromagnet
        bqm = dimod.BinaryQuadraticModel('SPIN')
        for u, v in itertools.combinations(chimera_sampler.largest_clique(), 2):
            bqm.quadratic[u, v] = -1

        chimera_sampler.sample(bqm).resolve()

    def test_too_large(self):
        num_variables = chimera_sampler.largest_clique_size + 1

        bqm = dimod.BinaryQuadraticModel(num_variables, 'SPIN')

        with self.assertRaises(ValueError):
            chimera_sampler.sample(bqm)

    def test_qubit_coupling_range(self):
        n = pegasus_sampler.largest_clique_size

        bqm = dimod.BinaryQuadraticModel({}, 
                {(u, v): -2 for u in range(n) for v in range(u+1, n)}, 'SPIN')

        pegasus_sampler.sample(bqm, chain_strength=-0.5).resolve()



class TestFailover(unittest.TestCase):

    @staticmethod
    def patch_solver(solver, exc):
        """Patch mock `solver` to fail with `exc` on sampleset resolve."""

        # clique sampler needs a valid topology
        solver.properties = {
            'topology': {
                'type': 'pegasus',
                'shape': [4]
            },
            'parameters': {},
            'h_range': [-2, 2],
            'j_range': [-1, 1]
        }

        # simulate solver failure on sampleset resolve
        fut = computation.Future(solver, None)
        fut._set_exception(exc)
        solver.sample_bqm = unittest.mock.Mock()
        solver.sample_bqm.return_value = fut

        return solver

    @unittest.mock.patch('dwave.system.samplers.dwave_sampler.Client')
    def test_default(self, mock_client):
        sampler = DWaveCliqueSampler()

        self.patch_solver(sampler.child.solver, exceptions.SolverOfflineError)

        with self.assertRaises(exceptions.SolverOfflineError):
            sampler.sample_ising({}, {}).resolve()

    @parameterized.expand([
        (exceptions.InvalidAPIResponseError, FailoverCondition),
        (exceptions.SolverNotFoundError, FailoverCondition),
        (exceptions.SolverOfflineError, FailoverCondition),
        (exceptions.SolverError, FailoverCondition),
        (exceptions.PollingTimeout, RetryCondition),
        (exceptions.SolverAuthenticationError, exceptions.SolverAuthenticationError),   # auth error propagated
        (KeyError, KeyError),   # unrelated errors propagated
    ])
    @unittest.mock.patch('dwave.system.samplers.dwave_sampler.Client')
    def test_async_failover(self, source_exc, target_exc, mock_client):
        sampler = DWaveCliqueSampler(failover=True)

        self.patch_solver(sampler.child.solver, source_exc)

        with self.assertRaises(target_exc):
            sampler.sample_ising({}, {}).resolve()

    @unittest.mock.patch('dwave.system.samplers.clique.DWaveSampler',
                         MockChimeraDWaveSampler)
    def test_properties_reinit(self):
        sampler = DWaveCliqueSampler()

        G = sampler.target_graph
        qlr = sampler.qpu_linear_range
        qqr = sampler.qpu_quadratic_range

        sampler.trigger_failover()

        self.assertIsNot(G, sampler.target_graph)
        self.assertIsNot(qlr, sampler.qpu_linear_range)
        self.assertIsNot(qqr, sampler.qpu_quadratic_range)



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from numbers import Number
from typing import Tuple

import warnings 

import dimod
import networkx as nx
import dwave_networkx as dnx

from minorminer.busclique import find_clique_embedding, busgraph_cache

try:
    from dwave.preprocessing import ScaleComposite
except ImportError:
    # fall back on dimod of dwave.preprocessing is not installed
    from dimod import ScaleComposite

from dwave.system.samplers.dwave_sampler import DWaveSampler
from dwave.system.coupling_groups import coupling_groups

__all__ = ['DWaveCliqueSampler']

class _QubitCouplingComposite(dimod.ComposedSampler):
    """Composite that scales variables of a problem.

    Checks whether the per qubit (or per group) coupling range is violated for
    the QPU and rescales accordingly. Scales the variables of a binary quadratic
    model (BQM) and modifies linear and quadratic terms accordingly.

    Args:
       sampler (:obj:`dimod.ComposedSampler`):
            A dimod sampler.
    """
    def __init__(self, child_sampler):
        self._children = [child_sampler]

    @property
    def children(self):
        return self._children

    @property
    def parameters(self):
        param = self.child.parameters.copy()
        return param

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    @dimod.decorators.nonblocking_sample_method
    def sample(self, bqm, **parameters):
        """ Scale and sample from the provided binary quadratic model.

        Problem is scaled based on the per qubit coupling range when 
        that range is exceeded. 

        Args:
            bqm (:obj:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            **parameters:
                Parameters for the sampling method, specified by the child
                sampler.

        Returns:
            :obj:`dimod.SampleSet`
        """
        if any(('per_qubit_coupling_range' in self.child.properties.keys(),
                'per_group_coupling_range' in self.child.properties.keys())):

            if 'per_qubit_coupling_range' in self.child.properties.keys():
                min_lim = self.child.properties['per_qubit_coupling_range'][0]
                max_lim = self.child.properties['per_qubit_coupling_range'][1]
                limit_name = 'per_qubit_coupling_range'

                total_coupling_range = {v: sum(bqm.adj[v].values())
                                        for v in bqm.variables}
            else:
                min_lim = self.child.properties['per_group_coupling_range'][0]
                max_lim = self.child.properties['per_group_coupling_range'][1]
                limit_name = 'per_group_coupling_range'

                hardware_graph = self.child.to_networkx_graph()
                total_coupling_range = {_: sum(bqm.quadratic.get(edge, 0) for edge in group)
                                        for _, group in enumerate(coupling_groups(hardware_graph))}

            min_coupling_range = min(total_coupling_range.values())
            max_coupling_range = max(total_coupling_range.values())

            if (min_coupling_range < min_lim or max_coupling_range > max_lim):
                warnings.warn(
                    f'The {limit_name} is violated after scaling.'
                    ' The problem is rescaled with respect to coupling range.'
                    ' No variables, interactions, or offset are ignored.')

                # scaling 
                inv_scalar = max(min_coupling_range / min_lim, 
                                 max_coupling_range / max_lim)
                scalar = 1.0 / inv_scalar

                bqm.scale(scalar,
                          ignored_variables=[],
                          ignored_interactions=[],
                          ignore_offset=[])

                sampleset = self.child.sample(bqm, **parameters)
                yield sampleset 

            else:
                sampleset = self.child.sample(bqm, **parameters)
                yield sampleset 
        else:
            sampleset = self.child.sample(bqm, **parameters)
            yield sampleset 

        yield sampleset 

class DWaveCliqueSampler(dimod.Sampler):
    """A sampler for solving clique binary quadratic models on the D-Wave system.

    This sampler wraps
    :func:`~minorminer.busclique.find_clique_embedding` to generate embeddings
    with even chain length. These embeddings work well for dense
    binary quadratic models. For sparse models, using
    :class:`.EmbeddingComposite` with :class:`.DWaveSampler` is preferred.

    Configuration such as :term:`solver` selection is similar to that of
    :class:`.DWaveSampler`.

    Args:
        failover (bool, optional, default=False):
            Signal a failover condition if a sampling error occurs. When ``True``,
            raises :exc:`~dwave.system.exceptions.FailoverCondition` or
            :exc:`~dwave.system.exceptions.RetryCondition` on sampleset resolve
            to signal failover.

            Actual failover, i.e. selection of a new solver, has to be handled
            by the user. A convenience method :meth:`.trigger_failover` is available
            for this. Note that hardware graphs vary between QPUs, so triggering
            failover results in regenerated :attr:`.nodelist`, :attr:`.edgelist`,
            :attr:`.properties` and :attr:`.parameters`.

            .. versionchanged:: 1.16.0

               In the past, the :meth:`.sample` method was blocking and
               ``failover=True`` caused a solver failover and sampling retry.
               However, this failover implementation broke when :meth:`sample`
               became non-blocking (asynchronous), Setting ``failover=True`` had
               no effect.

        retry_interval (number, optional, default=-1):
            Ignored, but kept for backward compatibility.

            .. versionchanged:: 1.16.0

               Ignored since 1.16.0. See note for ``failover`` parameter above.

        **config:
            Keyword arguments, as accepted by :class:`.DWaveSampler`

    Examples:
        This example creates a BQM based on a 6-node clique (complete graph),
        with random :math:`\pm 1` values assigned to nodes, and submits it to
        a D-Wave system. Parameters for communication with the system, such
        as its URL and an authentication token, are implicitly set in a
        configuration file or as environment variables, as described in
        `Configuring Access to D-Wave Solvers <https://docs.ocean.dwavesys.com/en/stable/overview/sapi.html>`_.

        >>> from dwave.system import DWaveCliqueSampler
        >>> import dimod
        ...
        >>> bqm = dimod.generators.ran_r(1, 6)
        ...
        >>> sampler = DWaveCliqueSampler()   # doctest: +SKIP
        >>> sampler.largest_clique_size > 5  # doctest: +SKIP
        True
        >>> sampleset = sampler.sample(bqm, num_reads=100)   # doctest: +SKIP

    """
    def __init__(self, *,
                 failover: bool = False, retry_interval: Number = -1,
                 **config):
        self.child = DWaveSampler(
            failover=failover, retry_interval=retry_interval, **config)

    @property
    def parameters(self) -> dict:
        try:
            return self._parameters
        except AttributeError:
            pass

        self._parameters = parameters = self.child.parameters.copy()

        # this sampler handles scaling
        parameters.pop('auto_scale', None)
        parameters.pop('bias_range', None)
        parameters.pop('quadratic_range', None)

        return parameters

    @property
    def properties(self) -> dict:
        try:
            return self._properties
        except AttributeError:
            pass

        self._properties = dict(qpu_properties=self.child.properties)
        return self.properties

    @property
    def largest_clique_size(self) -> int:
        """The maximum number of variables that can be embedded."""
        return len(self.largest_clique())

    @property
    def qpu_linear_range(self) -> Tuple[float, float]:
        """Range of linear biases allowed by the QPU."""
        try:
            return self._qpu_linear_range
        except AttributeError:
            pass

        # get the energy range
        try:
            energy_range = tuple(self.child.properties['h_range'])
        except KeyError as err:
            # for backwards compatibility with old software solvers
            if self.child.solver.software:
                energy_range = (-2, 2)
            else:
                raise err

        self._qpu_linear_range = energy_range

        return energy_range

    @property
    def qpu_quadratic_range(self) -> Tuple[float, float]:
        """Range of quadratic biases allowed by the QPU."""
        try:
            return self._qpu_quadratic_range
        except AttributeError:
            pass

        # get the energy range
        try:
            energy_range = tuple(
                self.child.properties.get('extended_j_range',
                                          self.child.properties['j_range']))
        except KeyError as err:
            # for backwards compatibility with old software solvers
            if self.child.solver.software:
                energy_range = (-1, 1)
            else:
                raise err

        self._qpu_quadratic_range = energy_range

        return energy_range

    @property
    def target_graph(self) -> nx.Graph:
        """The QPU topology."""
        try:
            return self._target_graph
        except AttributeError:
            pass

        self._target_graph = G = self.child.to_networkx_graph()

        return G

    def clique(self, variables):
        """Return a clique embedding of the given size.

        Args:
            variables (int/collection):
                Source variables. If an integer, the variables  embedded are
                labelled `[0,n)`.

        Returns:
            dict: The clique embedding.
        """
        return find_clique_embedding(variables, self.target_graph)

    def largest_clique(self):
        """The clique embedding with the maximum number of source variables.

        Returns:
            dict: The clique embedding with the maximum number of source
            variables.

        """
        return busgraph_cache(self.target_graph).largest_clique()

    def trigger_failover(self):
        """Trigger a failover and connect to a new solver."""

        self.child.trigger_failover()

        try:
            del self._target_graph
        except AttributeError:
            pass

        try:
            del self._qpu_linear_range
        except AttributeError:
            pass

        try:
            del self._qpu_quadratic_range
        except AttributeError:
            pass

    def sample(self, bqm, chain_strength=None, **kwargs):
        """Sample from the specified binary quadratic model.

        Args:
            bqm (:class:`~dimod.BinaryQuadraticModel`):
                Any binary quadratic model with up to
                :attr:`.largest_clique_size` variables. This BQM is embedded
                using a clique embedding.

            chain_strength (float/mapping/callable, optional):
                Sets the coupling strength between qubits representing variables
                that form a :term:`chain`. Mappings should specify the required
                chain strength for each variable. Callables should accept the BQM
                and embedding and return a float or mapping. By default,
                `chain_strength` is calculated with
                :func:`~dwave.embedding.chain_strength.uniform_torque_compensation`.

            **kwargs:
                Optional keyword arguments for the sampling method, specified
                per solver in :attr:`.parameters`.
                D-Wave System Documentation's
                `solver guide <https://docs.dwavesys.com/docs/latest/doc_solver_ref.html>`_
                describes the parameters and properties supported on the D-Wave
                system. Note that `auto_scale` is not supported by this
                sampler, because it scales the problem as part of the embedding
                process.

        Returns:
            :class:`~dimod.SampleSet`: Sample set constructed from a (non-blocking)
            :class:`~concurrent.futures.Future`-like object.

        """

        # some arguments should not be overwritten
        if 'auto_scale' in kwargs:
            raise TypeError("sample() got an unexpected keyword argument "
                            "'auto_scale'")
        if 'bias_range' in kwargs:
            raise TypeError("sample() got an unexpected keyword argument "
                            "'bias_range'")
        if 'quadratic_range' in kwargs:
            raise TypeError("sample() got an unexpected keyword argument "
                            "'quadratic_range'")

        # handle circular import. todo: fix
        from dwave.system.composites.embedding import FixedEmbeddingComposite

        # get the embedding
        embedding = find_clique_embedding(bqm.variables, self.target_graph,
                                          use_cache=True)

        # returns an empty embedding when the BQM is too large
        if not embedding and bqm.num_variables:
            raise ValueError("Cannot embed given BQM (size {}), sampler can "
                             "only handle problems of size {}".format(
                                len(bqm.variables), self.largest_clique_size))

        assert bqm.num_variables == len(embedding)  # sanity check

        # scaling only make sense in Ising space
        original_bqm = bqm

        if bqm.vartype is not dimod.SPIN:
            bqm = bqm.change_vartype(dimod.SPIN, inplace=False)

        sampler = FixedEmbeddingComposite(
            ScaleComposite(_QubitCouplingComposite(self.child)),
            embedding)

        if 'auto_scale' in self.child.parameters:
            kwargs['auto_scale'] = False

        sampleset = sampler.sample(bqm,
                                   bias_range=self.qpu_linear_range,
                                   quadratic_range=self.qpu_quadratic_range,
                                   chain_strength=chain_strength,
                                   **kwargs
                                   )

        # change_vartype is non-blocking
        return sampleset.change_vartype(original_bqm.vartype)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest

import dimod
import dimod.testing as dtest
import dwave_networkx
from dimod import ExactSolver

from dwave.system import ReverseBatchStatesComposite, ReverseAdvanceComposite

C4 = dwave_networkx.chimera_graph(4, 4, 4)


class MockReverseSampler(dimod.Sampler, dimod.Structured):
    nodelist = None
    edgelist = None
    properties = None
    parameters = None

    def __init__(self, broken_nodes=None):
        if broken_nodes is None:
            self.nodelist = sorted(C4.nodes)
            self.edgelist = sorted(sorted(edge) for edge in C4.edges)
        else:
            self.nodelist = sorted(v for v in C4.nodes if v not in broken_nodes)
            self.edgelist = sorted(sorted((u, v)) for u, v in C4.edges
                                   if u not in broken_nodes and v not in broken_nodes)

        # mark the sample kwargs
        self.parameters = parameters = {}
        parameters['num_reads'] = []
        parameters['initial_state'] = []
        parameters['reinitialize_state'] = []
        parameters['anneal_schedule'] = []

        # add the interesting properties manually
        self.properties = properties = {}
        properties['j_range'] = [-2.0, 1.0]
        properties['h_range'] = [-2.0, 2.0]
        properties['num_reads_range'] = [1, 10000]
        properties['num_qubits'] = len(C4)

    @dimod.bqm_structured
    def sample(self, bqm, **parameters):
        # we are altering the bqm
        return ExactSolver().sample(bqm)


class TestConstruction(unittest.TestCase):
    def test_instantiation_smoketest_advance(self):
        sampler = ReverseAdvanceComposite(dimod.ExactSolver())
        dtest.assert_sampler_api(sampler)

    def test_instantiation_smoketest_batch(self):
        sampler = ReverseBatchStatesComposite(dimod.ExactSolver())
        dtest.assert_sampler_api(sampler)


class TestReverseIsing(unittest.TestCase):
    def test_sample_ising_batch(self):
        sampler = ReverseBatchStatesComposite(MockReverseSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}

        response = sampler.sample_ising(h, J, initial_states=[{0: 1, 4: 1}, {0: -1, 4: -1}, {0: 1, 4: -1}])

        # nothing failed and we got at least three responses back for each initial state
        self.assertGreaterEqual(len(response), 3)

    def test_sample_ising_advance(self):
        sampler = ReverseAdvanceComposite(MockReverseSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}
        schedules = [[[0, 1], [1, 0.5], [2, 0.5], [3, 1]], [[0, 1], [1, 0.5], [2, 0.5], [3, 1]]]
        response = sampler.sample_ising(h, J, anneal_schedules=schedules)

        # nothing failed and we got at least two response back for each schedule point
        self.assertGreaterEqual(len(response), 2)

    def test_batch_correct_states(self):
        sampler = ReverseBatchStatesComposite(MockReverseSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}
        initial_states = [{0: 1, 4: 1}, {0: -1, 4: -1}, {0: 1, 4: -1}]
        response = sampler.sample_ising(h, J, initial_states=initial_states)

        variables = response.variables
        initial_state_list = response.record.initial_state

        for state in initial_states:
            state = [state[var] for var in variables]
            self.assertIn(state, initial_state_list)

    def test_batch_no_initial_states(self):
        sampler = ReverseBatchStatesComposite(MockReverseSampler())
        
        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}

        # check default generates an initial state
        response = sampler.sample_ising(h, J)
        self.assertIn('initial_state', response.record.dtype.names)

    def test_batch_generate_more_initial_states(self):
        sampler = ReverseBatchStatesComposite(MockReverseSampler())
        
        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}

        num_reads = 2
        initial_states = [{0:1, 4:1}]

        response = sampler.sample_ising(h, J, initial_states=initial_states, num_reads=num_reads)
        self.assertGreaterEqual(len(response), 4)

        response = sampler.sample_ising(h, J, initial_states=initial_states, initial_states_generator='tile', num_reads=num_reads)
        self.assertEqual(len(response), 4)  

        with self.assertRaises(ValueError):
            response = sampler.sample_ising(h, J, initial_states=initial_states, initial_states_generator='none', num_reads=num_reads)

    def test_batch_truncate_initial_states(self):
        sampler = ReverseBatchStatesComposite(MockReverseSampler())
        
        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}

        num_reads = 1
        initial_states = [{0:1, 4:1}, {0:-1, 4:1}, {0:-1, 4:-1}]

        response = sampler.sample_ising(h, J, initial_states=initial_states, num_reads=num_reads)
        self.assertEqual(len(response), 4)

        response = sampler.sample_ising(h, J, initial_states=initial_states, initial_states_generator='tile', num_reads=num_reads)
        self.assertEqual(len(response), 4)

        response = sampler.sample_ising(h, J, initial_states=initial_states, initial_states_generator='none', num_reads=num_reads)
        self.assertEqual(len(response), 4)

    def test_advance_no_schedules(self):
        sampler = ReverseAdvanceComposite(MockReverseSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}

        response = sampler.sample_ising(h, J)
        self.assertIn('schedule_index', response.record.dtype.names)

    def test_advance_correct_schedules(self):
        sampler = ReverseAdvanceComposite(MockReverseSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}
        anneal_schedules = [[[0, 1], [1, 0.5], [2, 0.5], [3, 1]], [[0, 1], [1, 0.5], [2, 0.5], [3, 1]]]
        response = sampler.sample_ising(h, J, anneal_schedules=anneal_schedules)

        anneal_schedule_list = response.record.schedule_index

        for schedule_idx in range(len(anneal_schedules)):
            self.assertIn(schedule_idx, anneal_schedule_list)

    def test_correct_initial_state_input(self):
        sampler = ReverseAdvanceComposite(MockReverseSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}
        anneal_schedules = [[[0, 1], [1, 0.5], [2, 0.5], [3, 1]], [[0, 1], [1, 0.5], [2, 0.5], [3, 1]]]

        with self.assertRaises(TypeError):
            response = sampler.sample_ising(h, J, anneal_schedules=anneal_schedules,
                                            initial_state=[1, 1])

    def test_correct_initial_state_used(self):
        sampler = ReverseAdvanceComposite(MockReverseSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}
        anneal_schedules = [[[0, 1], [1, 0.5], [2, 0.5], [3, 1]], [[0, 1], [1, 0.5], [2, 0.5], [3, 1]]]
        initial = {0: -1, 4: -1}
        response = sampler.sample_ising(h, J, anneal_schedules=anneal_schedules,
                                        initial_state=initial)

        vars = response.variables
        for datum in response.data(fields=['initial_state', 'schedule_index']):
            if datum.schedule_index == 0:
                self.assertListEqual([initial[v] for v in vars], list(datum.initial_state)) # initial_state = state that was passed in
            else:
                self.assertListEqual([1, -1], list(datum.initial_state)) # initial_state = best state found in last schedule

    def test_correct_initial_state_used_reinit(self):
        sampler = ReverseAdvanceComposite(MockReverseSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}
        anneal_schedules = [[[0, 1], [1, 0.5], [2, 0.5], [3, 1]], [[0, 1], [1, 0.5], [2, 0.5], [3, 1]]]
        initial = {0: -1, 4: -1}

        response = sampler.sample_ising(h, J, anneal_schedules=anneal_schedules,
                                        initial_state=initial, reinitialize_state=False)

        vars = response.variables

        init = [initial[v] for v in vars]  
        for datum in response.data(fields=['sample', 'initial_state'], sorted_by=None):
            self.assertListEqual(init, list(datum.initial_state))
            init = [datum.sample[v] for v in vars]  # sample should be the initial state of the next sample

    def test_combination(self):
        sampler = ReverseBatchStatesComposite(ReverseAdvanceComposite(MockReverseSampler()))

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}
        anneal_schedules = [[[0, 1], [1, 0.5], [2, 0.5], [3, 1]], [[0, 1], [1, 0.5], [2, 0.5], [3, 1]]]
        initial_states = [{0: 1, 4: 1}, {0: -1, 4: -1}, {0: 1, 4: -1}]
        response = sampler.sample_ising(h, J, anneal_schedules=anneal_schedules, initial_states=initial_states)

        anneal_schedule_list = response.record.schedule_index
        variables = response.variables
        initial_state_list = response.record.initial_state

        for state in initial_states:
            state = [state[var] for var in variables]
            self.assertIn(state, initial_state_list)

        for state_idx in range(len(anneal_schedules)):
            self.assertIn(state_idx, anneal_schedule_list)

        self.assertGreaterEqual(len(response), 6)


class TestReverseBinary(unittest.TestCase):
    def test_sample_qubo_batch(self):
        sampler = ReverseBatchStatesComposite(MockReverseSampler())

        Q = {(0, 4): 1.5}

        response = sampler.sample_qubo(Q, initial_states=[{0: 1, 4: 1}, {0: 0, 4: 0}, {0: 1, 4: 0}])

        # nothing failed and we got at least three responses back for each initial state
        self.assertGreaterEqual(len(response), 3)

    def test_sample_qubo_advance(self):
        sampler = ReverseAdvanceComposite(MockReverseSampler())

        h = {0: -1., 4: 2}
        J = {(0, 4): 1.5}
        schedules = [[[0, 1], [1, 0.5], [2, 0.5], [3, 1]], [[0, 1], [1, 0.5], [2, 0.5], [3, 1]]]
        response = sampler.sample_ising(h, J, schedules=schedules)

        # nothing failed and we got at least two response back for each schedule point
        self.assertGreaterEqual(len(response), 2)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import unittest

from dwave.cloud.exceptions import ConfigFileError, SolverNotFoundError

from dwave.system.schedules import ramp
from dwave.system.samplers import DWaveSampler


@unittest.skipIf(os.getenv('SKIP_INT_TESTS'), "Skipping integration test.")
class TestRamp(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        try:
            cls.qpu = DWaveSampler(solver=dict(h_gain_schedule=True))
        except (ValueError, ConfigFileError, SolverNotFoundError):
            raise unittest.SkipTest("no qpu available")

    @classmethod
    def tearDownClass(cls):
        cls.qpu.client.close()

    def test_with_h_gain_schedule(self):
        sampler = self.qpu

        schedule = ramp(.5, .2, sampler.properties['default_annealing_time'])

        sampler.validate_anneal_schedule(schedule)

        h = {v: 1 for v in sampler.nodelist}

        sampleset = sampler.sample_ising(h, {}, h_gain_schedule=schedule)
        sampleset.record  # resolve the future



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import unittest
import itertools
import random

from collections.abc import Mapping

import networkx as nx
import numpy as np
import numpy.testing as npt

import dimod
import dwave.embedding


class TestTargetToSource(unittest.TestCase):
    def test_identity_embedding(self):
        """a 1-to-1 embedding should not change the adjacency"""
        target_adj = nx.karate_club_graph()

        embedding = {v: {v} for v in target_adj}

        source_adj = dwave.embedding.target_to_source(target_adj, embedding)

        # test the adjacencies are equal (source_adj is a dict and target_adj is a networkx graph)
        for v in target_adj:
            self.assertIn(v, source_adj)
            for u in target_adj[v]:
                self.assertIn(u, source_adj[v])

        for v in source_adj:
            self.assertIn(v, target_adj)
            for u in source_adj[v]:
                self.assertIn(u, target_adj[v])

    def test_embedding_to_one_node(self):
        """an embedding that maps everything to one node should result in a singleton graph"""
        target_adj = nx.barbell_graph(16, 7)
        embedding = {'a': set(target_adj)}  # all map to 'a'

        source_adj = dwave.embedding.target_to_source(target_adj, embedding)
        self.assertEqual(source_adj, {'a': set()})

        embedding = {'a': {0, 1}}  # not every node is assigned to a chain
        source_adj = dwave.embedding.target_to_source(target_adj, embedding)
        self.assertEqual(source_adj, {'a': set()})

    def test_embedding_overlap(self):
        """overlapping embeddings should raise an error"""
        target_adj = nx.complete_graph(5)
        embedding = {'a': {0, 1}, 'b': {1, 2}}  # overlap

        with self.assertRaises(ValueError):
            source_adj = dwave.embedding.target_to_source(target_adj, embedding)

    def test_square_to_triangle(self):
        target_adjacency = {0: {1, 3}, 1: {0, 2}, 2: {1, 3}, 3: {0, 2}}  # a square graph
        embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        source_adjacency = dwave.embedding.target_to_source(target_adjacency, embedding)
        self.assertEqual(source_adjacency, {'a': {'b', 'c'}, 'b': {'a', 'c'}, 'c': {'a', 'b'}})


class TestEdgelistToAdjacency(unittest.TestCase):
    def test_typical(self):
        graph = nx.barbell_graph(17, 8)

        edgelist = set(graph.edges())

        adj = dwave.embedding.utils.edgelist_to_adjacency(edgelist)

        # test that they're equal
        for u, v in edgelist:
            self.assertIn(u, adj)
            self.assertIn(v, adj)
            self.assertIn(u, adj[v])
            self.assertIn(v, adj[u])

        for u in adj:
            for v in adj[u]:
                self.assertTrue((u, v) in edgelist or (v, u) in edgelist)
                self.assertFalse((u, v) in edgelist and (v, u) in edgelist)


class TestAdjacencyToEdgeIter(unittest.TestCase):
    def test_dict(self):
        graph = nx.barbell_graph(17, 8)
        adj = {v: set(graph[v]) for v in graph}

        edgelist = dwave.embedding.utils.adjacency_to_edges(adj)
        new_adj = {}
        for u, v in edgelist:
            new_adj.setdefault(u, set()).add(v)
            new_adj.setdefault(v, set()).add(u)

        self.assertEqual(adj, new_adj)

    def test_nxgraph(self):
        graph = nx.barbell_graph(17, 8)

        edgelist = dwave.embedding.utils.adjacency_to_edges(graph)

        edges0 = sorted(map(sorted, graph.edges()))
        edges1 = sorted(map(sorted, edgelist))

        self.assertEqual(edges0, edges1)

    def test_bqm(self):
        graph = nx.barbell_graph(17, 8)

        bqm = dimod.BQM(vartype = dimod.SPIN)

        bqm.add_interactions_from((u, v, 1) for u, v in graph.edges())

        edgelist = dwave.embedding.utils.adjacency_to_edges(bqm)
        
        edges0 = sorted(map(sorted, graph.edges()))
        edges1 = sorted(map(sorted, edgelist))

        self.assertEqual(edges0, edges1)

    def test_bad(self):
        with self.assertRaises(TypeError):
            edgelist = list(dwave.embedding.utils.adjacency_to_edges([]))

class TestChainToQuadratic(unittest.TestCase):
    def test_K5(self):
        """Test that when given a chain, the returned Jc uses all
        available edges."""
        chain_variables = set(range(5))

        # fully connected
        adjacency = {u: set(chain_variables) for u in chain_variables}
        for v, neighbors in adjacency.items():
            neighbors.remove(v)

        Jc = dwave.embedding.chain_to_quadratic(chain_variables, adjacency, 1.0)

        for u, v in itertools.combinations(chain_variables, 2):
            self.assertFalse((u, v) in Jc and (v, u) in Jc)
            self.assertTrue((u, v) in Jc or (v, u) in Jc)
        for u in chain_variables:
            self.assertFalse((u, u) in Jc)

    def test_5_cycle(self):
        chain_variables = set(range(5))

        # now try a cycle
        adjacency = {v: {(v + 1) % 5, (v - 1) % 5} for v in chain_variables}

        Jc = dwave.embedding.chain_to_quadratic(chain_variables, adjacency, 1.0)

        for u in adjacency:
            for v in adjacency[u]:
                self.assertFalse((u, v) in Jc and (v, u) in Jc)
                self.assertTrue((u, v) in Jc or (v, u) in Jc)

    def test_disconnected(self):
        chain_variables = {0, 2}

        adjacency = {0: {1}, 1: {0, 2}, 2: {1}}

        with self.assertRaises(ValueError):
            dwave.embedding.chain_to_quadratic(chain_variables, adjacency, 1.0)


class TestChainBreakFrequency(unittest.TestCase):
    def test_matrix_all_ones(self):
        """should have no breaks"""

        samples = np.ones((10, 5))

        embedding = {'a': {2, 4}, 'b': {1, 3}}

        freq = dwave.embedding.chain_break_frequency(samples, embedding)

        self.assertEqual(freq, {'a': 0, 'b': 0})

    def test_matrix_all_zeros(self):
        """should have no breaks"""

        samples = np.zeros((10, 5))

        embedding = {'a': {2, 4}, 'b': {1, 3}}

        freq = dwave.embedding.chain_break_frequency(samples, embedding)

        self.assertEqual(freq, {'a': 0, 'b': 0})

    def test_matrix_mix(self):
        samples = np.array([[-1, 1], [1, 1]])

        embedding = {'a': {0, 1}}

        freq = dwave.embedding.chain_break_frequency(samples, embedding)

        self.assertEqual(freq, {'a': .5})

    def test_mix_string_labels(self):
        response = dimod.SampleSet.from_samples([{'a': 1, 'b': 0}, {'a': 0, 'b': 0}],
                                                energy=[1, 0], info={}, vartype=dimod.BINARY)
        embedding = {0: {'a', 'b'}}
        freq = dwave.embedding.chain_break_frequency(response, embedding)

        self.assertEqual(freq, {0: .5})

    def test_with_num_occurences(self):
        samples = [[-1, -1, +1],
                   [-1, +1, +1],
                   [-1, +1, +1],
                   [-1, -1, -1],
                   [-1, +1, +1]]
        labels = 'abc'

        sampleset = dimod.SampleSet.from_samples((samples, labels), energy=0,
                                                 vartype=dimod.SPIN)
        sampleset = sampleset.aggregate()

        embedding = {0: 'ab', 1: 'c'}

        freq = dwave.embedding.chain_break_frequency(sampleset, embedding)

        self.assertEqual(freq, {0: 3./5, 1: 0})

class TestIntLabelDisjointSets(unittest.TestCase):
    def test(self):
        components = map(list, [range(1), range(1, 3), range(3, 6), range(6, 12)])
        djs = dwave.embedding.utils.intlabel_disjointsets(12)

        for x in range(12):
            self.assertEqual(djs.find(x), x)
            self.assertEqual(djs.size(x), 1)

        for c in components:
            for i, j in zip(c, c[1:]):
                djs.union(i, j)

        for c in components:
            root = djs.find(c[0])
            for x in c:
                self.assertEqual(djs.find(x), root)
                self.assertEqual(djs.size(x), len(c))

        for c in components:
            for i, j in combinations(c, 2):
                djs.union(i, j)

        for c in components:
            root = djs.find(c[0])
            for x in c:
                self.assertEqual(djs.find(x), root)
                self.assertEqual(djs.size(x), len(c))

        djs.union(0, 1)
        djs.union(2, 3)
        djs.union(5, 6)
        
        root = djs.find(0)
        for x in range(12):
            self.assertEqual(djs.find(x), root)
            self.assertEqual(djs.size(x), 12)
        





# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import collections.abc as abc
import itertools
import warnings

import numpy as np
import dimod

from collections import defaultdict
from copy import deepcopy

from dwave.embedding.chain_breaks import majority_vote, broken_chains
from dwave.embedding.exceptions import MissingEdgeError, MissingChainError, InvalidNodeError, DisconnectedChainError
from dwave.embedding.utils import adjacency_to_edges, intlabel_disjointsets
from dwave.embedding.chain_strength import uniform_torque_compensation


try:
    from dimod import AdjArrayBQM
    from dimod.core.bqm import BinaryView, SpinView
except ImportError:
    # dummy classes for isinstance checks
    class AdjArrayBQM: pass
    class BinaryView: pass
    class SpinView: pass


__all__ = ['embed_bqm',
           'embed_ising',
           'embed_qubo',
           'unembed_sampleset',
           'EmbeddedStructure',
           ]

class EmbeddedStructure(dict):
    """Processes an embedding and a target graph to collect target edges
    into those within individual chains, and those that connect chains.  This
    is used elsewhere to embed binary quadratic models into the target graph.

    Args:

        target_edges (iterable[edge]):
            An iterable of edges in the target graph.  Each edge should be an
            iterable of 2 hashable objects.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form
            {s: {t, ...}, ...}, where s is a source-model variable and t is a
            target-model variable.


    This class is a dict, and acts as an immutable duplicate of embedding.
    """

    def __init__(self, target_edges, embedding):
        if isinstance(embedding, EmbeddedStructure):
            super().__init__(embedding)
            if target_edges is None:
                # this condition is used by self.copy
                self._interaction_edges = embedding._interaction_edges.copy()
                self._chain_edges = embedding._chain_edges.copy()
                return
        else:
            super().__init__((u, tuple(c)) for u, c in embedding.items())

        target_label = {}
        self._chain_edges = chain_edges = {}
        self._interaction_edges = interaction_edges = defaultdict(list)
        self._chain_strength = None

        disjoint_sets = {}
        # prepare the data structures and compute the labeling of target nodes
        for u, emb_u in self.items():
            chain_edges[u] = []
            if not emb_u:
                raise MissingChainError(u)
            disjoint_sets[u] = intlabel_disjointsets(len(emb_u))
            for i, q in enumerate(emb_u):
                target_label[q] = u, i

        # filter the target edges into / between chain components
        for p, q in target_edges:
            if p in target_label and q in target_label:
                u, i = target_label[p]
                v, j = target_label[q]
                if u == v:
                    chain_edges[u].append((i, j))
                    disjoint_sets[u].union(i, j)
                else:
                    interaction_edges[u, v].append(i)
                    interaction_edges[v, u].append(j)

        for u, emb_u in self.items():
            if len(emb_u) != disjoint_sets[u].size(0):
                raise DisconnectedChainError(u)

    def __copy__(self):
        return EmbeddedStructure(None, self)

    copy = __copy__

    def __deepcopy__(self, memo):
        id_ = id(self)
        new = memo.get(id_, None)
        if new is not None:
            return new

        # everything is hashable so all copies are deep. Technically this
        # can break some object equivalence (because we're ignoring the
        # memo) but since the chains are converted into a new tuple in
        # the __init__ already... this seems simpler and cleaner.
        new = memo[id_] = self.copy()
        return new

    @property
    def chain_strength(self):
        """callable: Return the chain strength selected for embedding."""
        return self._chain_strength

    def chain_edges(self, u):
        """Iterate over edges contained in the chain for u.

        Args:
            u (hashable):
                A key in self.

        Yields:
            tuple: A 2-tuple, corresponding to an edge in the target graph.

        """
        emb_u = self[u]
        for i, j in self._chain_edges[u]:
            yield emb_u[i], emb_u[j]


    def interaction_edges(self, u, *args):
        """Iterate over edges between in the chains for u and v.

        Args:
            u (hashable/tuple):
                A key in self.

            v (hashable, optional):
                A key in self.  If this argument is omitted, interprets
                :code:`u, v := u`.

        Yields:
            tuple: A 2-tuple, corresponding to an edge in the target graph.

        """
        if args:
            v, = args
        else:
            u, v = u
        emb_u = self[u]
        emb_v = self[v]
        int_u = self._interaction_edges[u, v]
        int_v = self._interaction_edges[v, u]
        for i, j in zip(int_u, int_v):
            yield emb_u[i], emb_v[j]

    def _mutate_dict(self, *a, **k):
        """Raise a TypeError -- this method is not supported because
        EmbeddedStructure is immutable, but exists because dict is the parent
        class."""
        raise TypeError("EmbeddedStructure is immutable")

    __delitem__=__setitem__=clear=pop=popitem=setdefault=update=_mutate_dict

    def fromkeys(self, *args, **kwargs):
        """Raise a NotImplemented -- this method is not supported for the
        EmbeddedStructure class, but exists because dict is the parent class."""
        raise NotImplementedError("EmbeddedStructure does not support the"
                                  " fromkeys method")

    def embed_bqm(self, source_bqm, chain_strength=None, smear_vartype=None):
        """Embed a binary quadratic model onto a target graph.

        Args:
            source_bqm (:class:`~dimod.BinaryQuadraticModel`):
                Binary quadratic model to embed.

            chain_strength (float/mapping/callable, optional):
                Sets the coupling strength between qubits representing variables 
                that form a :term:`chain`. Mappings should specify the required 
                chain strength for each variable. Callables should accept the BQM 
                and embedding and return a float or mapping. By default, 
                `chain_strength` is calculated with
                :func:`~dwave.embedding.chain_strength.uniform_torque_compensation`.

            smear_vartype (:class:`.Vartype`, optional, default=None):
                Determines whether the linear bias of embedded variables is
                smeared (the specified value is evenly divided as biases of a
                chain in the target graph) in SPIN or BINARY space. Defaults to
                the :class:`.Vartype` of `source_bqm`.

        Returns:
            :obj:`.BinaryQuadraticModel`: Target binary quadratic model.

        Examples:
            This example embeds a triangular binary quadratic model representing
            a :math:`K_3` clique into a square target graph by mapping variable
            `c` in the source to nodes `2` and `3` in the target.

            >>> import networkx as nx
            ...
            >>> target = nx.cycle_graph(4)
            >>> # Binary quadratic model for a triangular source graph
            >>> h = {'a': 0, 'b': 0, 'c': 0}
            >>> J = {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1}
            >>> bqm = dimod.BinaryQuadraticModel.from_ising(h, J)
            >>> # Variable c is a chain
            >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
            >>> # Embed and show the chain strength
            >>> target_bqm = dwave.embedding.embed_bqm(bqm, embedding, target)
            >>> target_bqm.quadratic[(2, 3)]
            -1.9996979771955565
            >>> print(target_bqm.quadratic)  # doctest: +SKIP
            {(0, 1): 1.0, (0, 3): 1.0, (1, 2): 1.0, (2, 3): -1.9996979771955565}


        See also:
            :func:`.embed_ising`, :func:`.embed_qubo`

        """
        return_vartype = source_bqm.vartype
        if smear_vartype is dimod.SPIN:
            source_bqm = source_bqm.spin
        elif smear_vartype is dimod.BINARY:
            source_bqm = source_bqm.binary
        else:
            smear_vartype = source_bqm.vartype

        # we need this check to support dimod 0.9.x, where there was one bqm
        # type that was not shapeable
        if (isinstance(source_bqm, (AdjArrayBQM, SpinView, BinaryView)) and
                not source_bqm.shapeable()):
            # this will never happen in dimod>=0.10.0
            target_bqm = dimod.AdjVectorBQM.empty(smear_vartype)
        else:
            try:
                # dimod 0.9.x
                target_bqm = source_bqm.base.empty(smear_vartype)
            except AttributeError:
                target_bqm = source_bqm.empty(smear_vartype)

        # add the offset
        target_bqm.offset += source_bqm.offset

        if chain_strength is None:
            chain_strength = uniform_torque_compensation

        if callable(chain_strength):
            chain_strength = chain_strength(source_bqm, self)

        self._chain_strength = chain_strength

        if isinstance(chain_strength, abc.Mapping):
            strength_iter = (chain_strength[v] for v in source_bqm.linear)
        else:
            strength_iter = itertools.repeat(chain_strength)

        offset = 0
        # spread the linear source bias equally over the target variables in the
        # chain and add chain edges as necessary
        for (v, bias), strength in zip(source_bqm.linear.items(), strength_iter):
            chain = self.get(v)
            if chain is None:
                raise MissingChainError(v)

            #we check that this is nonzero in __init__
            b = bias / len(chain)

            target_bqm.add_variables_from({u: b for u in chain})

            if len(chain) == 1:
                # in the case where the chain has length 1, there are no chain
                # quadratic biases, but we none-the-less want the chain
                # variables to appear in the target_bqm
                q, = chain
                target_bqm.add_variable(q, 0.0)
            elif smear_vartype is dimod.SPIN:
                for p, q in self.chain_edges(v):
                    target_bqm.add_interaction(p, q, -strength)
                    offset += strength
            else:
                # this is in spin, but we need to respect the vartype
                for p, q in self.chain_edges(v):
                    target_bqm.add_interaction(p, q, -4*strength)
                    target_bqm.add_variable(p, 2*strength)
                    target_bqm.add_variable(q, 2*strength)

        target_bqm.offset += offset

        # next up the quadratic biases, spread the quadratic biases evenly over
        # the available interactions
        for (u, v), bias in source_bqm.quadratic.items():
            interactions = list(self.interaction_edges(u, v))

            if not interactions:
                raise MissingEdgeError(u, v)

            b = bias / len(interactions)

            target_bqm.add_interactions_from((u, v, b) for u, v in interactions)

        # we made the target BQM so we can safely mutate it in-place
        return target_bqm.change_vartype(return_vartype, inplace=True)


def embed_bqm(source_bqm, embedding=None, target_adjacency=None,
              chain_strength=None, smear_vartype=None):
    """Embed a binary quadratic model onto a target graph.

    Args:
        source_bqm (:class:`~dimod.BinaryQuadraticModel`):
            Binary quadratic model to embed.

        embedding (dict/:class:`.EmbeddedStructure`):
            Mapping from source graph to target graph as a dict of form
            {s: {t, ...}, ...}, where s is a source-model variable and t is a
            target-model variable.  Alternately, an EmbeddedStructure object
            produced by, for example,
            EmbeddedStructure(target_adjacency.edges(), embedding). If embedding
            is a dict, target_adjacency must be provided.

        target_adjacency (dict/:obj:`networkx.Graph`, optional):
            Adjacency of the target graph as a dict of form {t: Nt, ...}, where
            t is a variable in the target graph and Nt is its set of neighbours.
            This should be omitted if and only if embedding is an
            EmbeddedStructure object.

        chain_strength (float/mapping/callable, optional):
            Sets the coupling strength between qubits representing variables 
            that form a :term:`chain`. Mappings should specify the required chain
            strength for each variable. Callables should accept the BQM and 
            embedding and return a float or mapping. By default, 
            `chain_strength` is calculated with
            :func:`~dwave.embedding.chain_strength.uniform_torque_compensation`.

        smear_vartype (:class:`.Vartype`, optional, default=None):
            Determines whether the linear bias of embedded variables is smeared
            (the specified value is evenly divided as biases of a chain in the
            target graph) in SPIN or BINARY space. Defaults to the
            :class:`.Vartype` of `source_bqm`.

    Returns:
        :obj:`.BinaryQuadraticModel`: Target binary quadratic model.

    Examples:
        This example embeds a triangular binary quadratic model representing
        a :math:`K_3` clique into a square target graph by mapping variable `c`
        in the source to nodes `2` and `3` in the target.

        >>> import networkx as nx
        ...
        >>> target = nx.cycle_graph(4)
        >>> # Binary quadratic model for a triangular source graph
        >>> h = {'a': 0, 'b': 0, 'c': 0}
        >>> J = {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1}
        >>> bqm = dimod.BinaryQuadraticModel.from_ising(h, J)
        >>> # Variable c is a chain
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> # Embed and show the chain strength
        >>> target_bqm = dwave.embedding.embed_bqm(bqm, embedding, target)
        >>> target_bqm.quadratic[(2, 3)]
        -1.9996979771955565
        >>> print(target_bqm.quadratic)  # doctest: +SKIP
        {(0, 1): 1.0, (0, 3): 1.0, (1, 2): 1.0, (2, 3): -1.9996979771955565}


    See also:
        :func:`.embed_ising`, :func:`.embed_qubo`

    """
    if isinstance(embedding, EmbeddedStructure):
        if target_adjacency is not None:
            warnings.warn(
                "target_adjacency should not be provided if embedding is an "
                "EmbeddedStructure. The given value will be ignored. In the "
                "future this will raise an exception",
                DeprecationWarning, stacklevel=2
                )
    elif target_adjacency is None:
        raise ValueError("either embedding should be an EmbeddedStructure, or "
                         "target_adjacency must be provided")
    else:
        target_edges = adjacency_to_edges(target_adjacency)
        embedding = EmbeddedStructure(target_edges, embedding)

    return embedding.embed_bqm(source_bqm, smear_vartype=smear_vartype,
                               chain_strength=chain_strength)


def embed_ising(source_h, source_J, embedding, target_adjacency, chain_strength=None):
    """Embed an Ising problem onto a target graph.

    Args:
        source_h (dict[variable, bias]/list[bias]):
            Linear biases of the Ising problem. If a list, the list's indices are used as
            variable labels.

        source_J (dict[(variable, variable), bias]):
            Quadratic biases of the Ising problem.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form {s: {t, ...}, ...},
            where s is a source-model variable and t is a target-model variable.

        target_adjacency (dict/:obj:`networkx.Graph`):
            Adjacency of the target graph as a dict of form {t: Nt, ...},
            where t is a target-graph variable and Nt is its set of neighbours.

        chain_strength (float/mapping/callable, optional):
            Sets the coupling strength between qubits representing variables 
            that form a :term:`chain`. Mappings should specify the required chain
            strength for each variable. Callables should accept the BQM and 
            embedding and return a float or mapping. By default, 
            `chain_strength` is calculated with
            :func:`~dwave.embedding.chain_strength.uniform_torque_compensation`.

    Returns:
        tuple: A 2-tuple:

            dict[variable, bias]: Linear biases of the target Ising problem.

            dict[(variable, variable), bias]: Quadratic biases of the target Ising problem.

    Examples:
        This example embeds a triangular Ising problem representing
        a :math:`K_3` clique into a square target graph by mapping variable `c`
        in the source to nodes `2` and `3` in the target.

        >>> import networkx as nx
        ...
        >>> target = nx.cycle_graph(4)
        >>> # Ising problem biases
        >>> h = {'a': 0, 'b': 0, 'c': 0}
        >>> J = {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1}
        >>> # Variable c is a chain
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> # Embed and show the resulting biases
        >>> th, tJ = dwave.embedding.embed_ising(h, J, embedding, target)
        >>> th  # doctest: +SKIP
        {0: 0.0, 1: 0.0, 2: 0.0, 3: 0.0}
        >>> tJ  # doctest: +SKIP
        {(0, 1): 1.0, (0, 3): 1.0, (1, 2): 1.0, (2, 3): -1.0}


    See also:
        :func:`.embed_bqm`, :func:`.embed_qubo`

    """
    source_bqm = dimod.BinaryQuadraticModel.from_ising(source_h, source_J)
    target_bqm = embed_bqm(source_bqm, embedding, target_adjacency, chain_strength=chain_strength)
    target_h, target_J, __ = target_bqm.to_ising()
    return target_h, target_J


def embed_qubo(source_Q, embedding, target_adjacency, chain_strength=None):
    """Embed a QUBO onto a target graph.

    Args:
        source_Q (dict[(variable, variable), bias]):
            Coefficients of a quadratic unconstrained binary optimization (QUBO) model.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form {s: {t, ...}, ...},
            where s is a source-model variable and t is a target-model variable.

        target_adjacency (dict/:obj:`networkx.Graph`):
            Adjacency of the target graph as a dict of form {t: Nt, ...},
            where t is a target-graph variable and Nt is its set of neighbours.

        chain_strength (float/mapping/callable, optional):
            Sets the coupling strength between qubits representing variables 
            that form a :term:`chain`. Mappings should specify the required chain
            strength for each variable. Callables should accept the BQM and 
            embedding and return a float or mapping. By default, 
            `chain_strength` is calculated with
            :func:`~dwave.embedding.chain_strength.uniform_torque_compensation`.

    Returns:
        dict[(variable, variable), bias]: Quadratic biases of the target QUBO.

    Examples:
        This example embeds a triangular QUBO representing a :math:`K_3` clique
        into a square target graph by mapping variable `c` in the source to nodes
        `2` and `3` in the target.

        >>> import networkx as nx
        ...
        >>> target = nx.cycle_graph(4)
        >>> # QUBO
        >>> Q = {('a', 'b'): 1, ('b', 'c'): 1, ('a', 'c'): 1}
        >>> # Variable c is a chain
        >>> embedding = {'a': {0}, 'b': {1}, 'c': {2, 3}}
        >>> # Embed and show the resulting biases
        >>> tQ = dwave.embedding.embed_qubo(Q, embedding, target)
        >>> tQ  # doctest: +SKIP
        {(0, 1): 1.0,
         (0, 3): 1.0,
         (1, 2): 1.0,
         (2, 3): -4.0,
         (0, 0): 0.0,
         (1, 1): 0.0,
         (2, 2): 2.0,
         (3, 3): 2.0}

    See also:
        :func:`.embed_bqm`, :func:`.embed_ising`

    """
    source_bqm = dimod.BinaryQuadraticModel.from_qubo(source_Q)
    target_bqm = embed_bqm(source_bqm, embedding, target_adjacency, chain_strength=chain_strength)
    target_Q, __ = target_bqm.to_qubo()
    return target_Q


def unembed_sampleset(target_sampleset, embedding, source_bqm,
                      chain_break_method=None, chain_break_fraction=False,
                      return_embedding=False):
    """Unembed a sample set.

    Given samples from a target binary quadratic model (BQM), construct a sample
    set for a source BQM by unembedding.

    Args:
        target_sampleset (:obj:`dimod.SampleSet`):
            Sample set from the target BQM.

        embedding (dict):
            Mapping from source graph to target graph as a dict of form
            {s: {t, ...}, ...}, where s is a source variable and t is a target
            variable.

        source_bqm (:obj:`~dimod.BinaryQuadraticModel`):
            Source BQM.

        chain_break_method (function/list, optional):
            Method or methods used to resolve chain breaks. If multiple methods
            are given, the results are concatenated and a new field called
            "chain_break_method" specifying the index of the method is appended
            to the sample set.
            Defaults to :func:`~dwave.embedding.chain_breaks.majority_vote`.
            See :mod:`dwave.embedding.chain_breaks`.

        chain_break_fraction (bool, optional, default=False):
            Add a `chain_break_fraction` field to the unembedded :obj:`dimod.SampleSet`
            with the fraction of chains broken before unembedding.

        return_embedding (bool, optional, default=False):
            If True, the embedding is added to :attr:`dimod.SampleSet.info`
            of the returned sample set. Note that if an `embedding` key
            already exists in the sample set then it is overwritten.

    Returns:
        :obj:`.SampleSet`: Sample set in the source BQM.

    Examples:
       This example unembeds from a square target graph samples of a triangular
       source BQM.

        >>> # Triangular binary quadratic model and an embedding
        >>> J = {('a', 'b'): -1, ('b', 'c'): -1, ('a', 'c'): -1}
        >>> bqm = dimod.BinaryQuadraticModel.from_ising({}, J)
        >>> embedding = {'a': [0, 1], 'b': [2], 'c': [3]}
        >>> # Samples from the embedded binary quadratic model
        >>> samples = [{0: -1, 1: -1, 2: -1, 3: -1},  # [0, 1] is unbroken
        ...            {0: -1, 1: +1, 2: +1, 3: +1}]  # [0, 1] is broken
        >>> energies = [-3, 1]
        >>> embedded = dimod.SampleSet.from_samples(samples, dimod.SPIN, energies)
        >>> # Unembed
        >>> samples = dwave.embedding.unembed_sampleset(embedded, embedding, bqm)
        >>> samples.record.sample   # doctest: +SKIP
        array([[-1, -1, -1],
               [ 1,  1,  1]], dtype=int8)

    """

    if chain_break_method is None:
        chain_break_method = majority_vote
    elif isinstance(chain_break_method, abc.Sequence):
        # we want to apply multiple CBM and then combine
        samplesets = [unembed_sampleset(target_sampleset, embedding,
                                        source_bqm, chain_break_method=cbm,
                                        chain_break_fraction=chain_break_fraction)
                      for cbm in chain_break_method]
        sampleset = dimod.sampleset.concatenate(samplesets)

        # Add a new data field tracking which came from
        # todo: add this functionality to dimod
        cbm_idxs = np.empty(len(sampleset), dtype=int)

        start = 0
        for i, ss in enumerate(samplesets):
            cbm_idxs[start:start+len(ss)] = i
            start += len(ss)

        new = np.lib.recfunctions.append_fields(sampleset.record,
                                                'chain_break_method', cbm_idxs,
                                                asrecarray=True, usemask=False)

        return type(sampleset)(new, sampleset.variables, sampleset.info,
                               sampleset.vartype)

    variables = list(source_bqm.variables)  # need this ordered
    try:
        chains = [embedding[v] for v in variables]
    except KeyError:
        raise ValueError("given bqm does not match the embedding")

    record = target_sampleset.record

    unembedded, idxs = chain_break_method(target_sampleset, chains)

    reserved = {'sample', 'energy'}
    vectors = {name: record[name][idxs]
               for name in record.dtype.names if name not in reserved}

    if chain_break_fraction:
        broken = broken_chains(target_sampleset, chains)
        if broken.size:
            vectors['chain_break_fraction'] = broken.mean(axis=1)[idxs]
        else:
            vectors['chain_break_fraction'] = 0

    info = target_sampleset.info.copy()

    if return_embedding:
        embedding_context = dict(embedding=embedding,
                                 chain_break_method=chain_break_method.__name__)
        info.update(embedding_context=embedding_context)

    return dimod.SampleSet.from_samples_bqm((unembedded, variables),
                                            source_bqm,
                                            info=info,
                                            **vectors)



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Composites that remove any interactions below a cutoff value. Isolated
variables are then also removed.
"""
import operator

import numpy as np

import dimod

__all__ = 'CutOffComposite', 'PolyCutOffComposite'


class CutOffComposite(dimod.ComposedSampler):
    """Composite to remove interactions below a specified cutoff value.

    Prunes the binary quadratic model (BQM) submitted to the child sampler by
    retaining only interactions with values commensurate with the sampler's
    precision as specified by the ``cutoff`` argument. Also removes variables
    isolated post- or pre-removal of these interactions from the BQM passed
    on to the child sampler, setting these variables to values that minimize
    the original BQM's energy for the returned samples.

    Args:
       sampler (:obj:`dimod.Sampler`):
            A dimod sampler.

       cutoff (number):
            Lower bound for absolute value of interactions. Interactions
            with absolute values lower than ``cutoff`` are removed. Isolated variables
            are also not passed on to the child sampler.

       cutoff_vartype (:class:`.Vartype`/str/set, default='SPIN'):
            Variable space to execute the removal in. Accepted input values:

            * :class:`.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
            * :class:`.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

       comparison (function, optional):
            A comparison operator for comparing interaction values to the cutoff
            value. Defaults to :func:`operator.lt`.

    Examples:
        This example removes one interaction, ``'ac': -0.7``, before embedding
        on a D-Wave system. Note that the lowest-energy sample for the embedded problem
        is unchanged ``{'a': 1, 'b': -1, 'c': -1}`` and this solution is found.
        However, the sample is attributed the energy appropriate to the bqm
        without thresholding.

        >>> import dimod
        >>> sampler = DWaveSampler(solver={'qpu': True})
        >>> bqm = dimod.BinaryQuadraticModel({'a': -1, 'b': 1, 'c': 1},
        ...                            {'ab': 0.8, 'ac': 0.7, 'bc': -1},
        ...                            0,
        ...                            dimod.SPIN)
        >>> CutOffComposite(AutoEmbeddingComposite(sampler), 0.75).sample(bqm,
        ...                 num_reads=1000).first.energy
        -5.5

    """

    @dimod.decorators.vartype_argument('cutoff_vartype')
    def __init__(self, child_sampler, cutoff, cutoff_vartype=dimod.SPIN,
                 comparison=operator.lt):
        self._children = [child_sampler]
        self._cutoff = cutoff
        self._cutoff_vartype = cutoff_vartype
        self._comparison = comparison

    @property
    def children(self):
        """List of child samplers that that are used by this composite."""
        return self._children

    @property
    def parameters(self):
        """A dict where keys are the keyword parameters accepted by the sampler methods
        and values are lists of the properties relevent to each parameter."""
        return self.child.parameters.copy()

    @property
    def properties(self):
        """A dict containing any additional information about the sampler."""
        return {'child_properties': self.child.properties.copy()}

    def sample(self, bqm, **parameters):
        """Cut off interactions and sample from the provided binary quadratic model.

        Prunes the binary quadratic model (BQM) submitted to the child sampler
        by retaining only interactions with value commensurate with the
        sampler's precision as specified by the ``cutoff`` argument. Also removes
        variables isolated post- or pre-removal of these interactions from the
        BQM passed on to the child sampler, setting these variables to values
        that minimize the original BQM's energy for the returned samples.

        Args:
            bqm (:obj:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            **parameters:
                Parameters for the sampling method, specified by the child sampler.

        Returns:
            :obj:`dimod.SampleSet`

        Examples:
            See the example in :class:`CutOffComposite`.

        """
        child = self.child
        cutoff = self._cutoff
        cutoff_vartype = self._cutoff_vartype
        comp = self._comparison

        if cutoff_vartype is dimod.SPIN:
            original = bqm.spin
        else:
            original = bqm.binary

        # remove all of the interactions less than cutoff
        new = type(bqm)(original.linear,
                        ((u, v, bias)
                         for (u, v), bias in original.quadratic.items()
                         if not comp(abs(bias), cutoff)),
                        original.offset,
                        original.vartype)

        # next we check for isolated qubits and remove them, we could do this as
        # part of the construction but the assumption is there should not be
        # a large number in the 'typical' case
        isolated = [v for v in new.variables if not new.adj[v]]
        new.remove_variables_from(isolated)

        if isolated and len(new) == 0:
            # in this case all variables are isolated, so we just put one back
            # to serve as the basis
            v = isolated.pop()
            new.linear[v] = original.linear[v]

        # get the samples from the child sampler and put them into the original vartype
        sampleset = child.sample(new, **parameters).change_vartype(bqm.vartype, inplace=True)

        # we now need to add the isolated back in, in a way that minimizes
        # the energy. There are lots of ways to do this but for now we'll just
        # do one
        if isolated:
            samples, variables = _restore_isolated(sampleset, bqm, isolated)
        else:
            samples = sampleset.record.sample
            variables = sampleset.variables

        vectors = sampleset.data_vectors
        vectors.pop('energy')  # we're going to recalculate the energy anyway

        return dimod.SampleSet.from_samples_bqm((samples, variables), bqm, **vectors)


def _restore_isolated(sampleset, bqm, isolated):
    """Return samples-like by adding isolated variables into sampleset in a
    way that minimizes the energy (relative to the other non-isolated variables).
    """

    samples = sampleset.record.sample
    variables = sampleset.variables

    new_samples = np.empty((len(sampleset), len(isolated)), dtype=samples.dtype)

    # we don't let the isolated variables interact with each other for now because
    # it will slow this down substantially
    for col, v in enumerate(isolated):
        try:
            neighbours, biases = zip(*((u, bias) for u, bias in bqm.adj[v].items()
                                       if u in variables))  # ignore other isolates
        except ValueError:
            # happens when only neighbors are other isolated variables
            new_samples[:, col] = bqm.linear[v] <= 0
            continue

        idxs = [variables.index(u) for u in neighbours]

        # figure out which value for v would minimize the energy
        # v(h_v + \sum_u J_uv * u)
        new_samples[:, col] = samples[:, idxs].dot(biases) < -bqm.linear[v]

    if bqm.vartype is dimod.SPIN:
        new_samples = 2*new_samples - 1

    return np.concatenate((samples, new_samples), axis=1), list(variables) + isolated


class PolyCutOffComposite(dimod.ComposedPolySampler):
    """Composite to remove polynomial interactions below a specified cutoff value.

    Prunes the binary polynomial submitted to the child sampler by retaining
    only interactions with values commensurate with the sampler's precision as
    specified by the ``cutoff`` argument. Also removes variables isolated post-
    or pre-removal of these interactions from the polynomial passed on to the
    child sampler, setting these variables to values that minimize the
    original polynomial's energy for the returned samples.

    Args:
       sampler (:obj:`dimod.PolySampler`):
            A dimod binary polynomial sampler.

       cutoff (number):
            Lower bound for absolute value of interactions. Interactions
            with absolute values lower than ``cutoff`` are removed. Isolated variables
            are also not passed on to the child sampler.

       cutoff_vartype (:class:`.Vartype`/str/set, default='SPIN'):
            Variable space to do the cutoff in. Accepted input values:

            * :class:`.Vartype.SPIN`, ``'SPIN'``, ``{-1, 1}``
            * :class:`.Vartype.BINARY`, ``'BINARY'``, ``{0, 1}``

       comparison (function, optional):
            A comparison operator for comparing the interaction value to the cutoff
            value. Defaults to :func:`operator.lt`.

    Examples:
        This example removes one interaction, ``'ac': 0.2``, before submitting
        the polynomial to child sampler :class:`~dimod.reference.samplers.ExactSolver`.

        >>> import dimod
        >>> sampler = dimod.HigherOrderComposite(dimod.ExactSolver())
        >>> poly = dimod.BinaryPolynomial({'a': 3, 'abc':-4, 'ac': 0.2}, dimod.SPIN)
        >>> PolyCutOffComposite(sampler, 1).sample_poly(poly).first.sample['a']
        -1

    """
    @dimod.decorators.vartype_argument('cutoff_vartype')
    def __init__(self, child_sampler, cutoff, cutoff_vartype=dimod.SPIN,
                 comparison=operator.lt):
        if not isinstance(child_sampler, dimod.PolySampler):
            raise TypeError("Child sampler must be a PolySampler")
        self._children = [child_sampler]
        self._cutoff = cutoff
        self._cutoff_vartype = cutoff_vartype
        self._comparison = comparison

    @property
    def children(self):
        """List of child samplers that that are used by this composite."""
        return self._children

    @property
    def parameters(self):
        """A dict where keys are the keyword parameters accepted by the sampler methods and values are lists of the properties relevent to each parameter."""
        return self.child.parameters.copy()

    @property
    def properties(self):
        """A dict containing any additional information about the sampler."""
        return {'child_properties': self.child.properties.copy()}

    def sample_poly(self, poly, **kwargs):
        """Cutoff and sample from the provided binary polynomial.

        Prunes the binary polynomial submitted to the child sampler by retaining
        only interactions with values commensurate with the sampler's precision
        as specified by the ``cutoff`` argument. Also removes variables isolated
        post- or pre-removal of these interactions from the polynomial passed
        on to the child sampler, setting these variables to values that minimize
        the original polynomial's energy for the returned samples.

        Args:
            poly (:obj:`dimod.BinaryPolynomial`):
                Binary polynomial to be sampled from.

            **parameters:
                Parameters for the sampling method, specified by the child sampler.

        Returns:
            :obj:`dimod.SampleSet`

        Examples:
            See the example in :class:`PolyCutOffComposite`.

        """
        child = self.child
        cutoff = self._cutoff
        cutoff_vartype = self._cutoff_vartype
        comp = self._comparison

        if cutoff_vartype is dimod.SPIN:
            original = poly.to_spin(copy=False)
        else:
            original = poly.to_binary(copy=False)

        # remove all of the terms of order >= 2 that have a bias less than cutoff
        new = type(poly)(((term, bias) for term, bias in original.items()
                          if len(term) > 1 and not comp(abs(bias), cutoff)),
                         cutoff_vartype)

        # also include the linear biases for the variables in new
        for v in new.variables:
            term = v,
            if term in original:
                new[term] = original[term]

        # everything else is isolated
        isolated = list(original.variables.difference(new.variables))

        if isolated and len(new) == 0:
            # in this case all variables are isolated, so find the variable with
            # the strongest bias and use that as the seed for putting the other
            # variables back in
            v = max(isolated, key=lambda v: original.get((v,), 0.0))
            isolated.remove(v)
            new[(v,)] = original.get((v,), 0)

        # get the samples from the child sampler and put them into the original vartype
        sampleset = child.sample_poly(new, **kwargs).change_vartype(poly.vartype, inplace=True)

        # we now need to add the isolated back in, in a way that minimizes
        # the energy. There are lots of ways to do this but for now we'll just
        # do one
        if isolated:
            samples, variables = _restore_isolated_higherorder(sampleset, poly, isolated)
        else:
            samples = sampleset.record.sample
            variables = sampleset.variables

        vectors = sampleset.data_vectors
        vectors.pop('energy')  # we're going to recalculate the energy anyway

        return dimod.SampleSet.from_samples_bqm((samples, variables), poly, **vectors)


def _restore_isolated_higherorder(sampleset, poly, isolated):
    """Return samples-like by adding isolated variables into sampleset in a
    way that minimizes the energy (relative to the other non-isolated variables).

    Isolated should be ordered.
    """

    samples = sampleset.record.sample
    variables = sampleset.variables

    new_samples = np.empty((len(sampleset), len(isolated)), dtype=samples.dtype)

    # we don't let the isolated variables interact with eachother for now because
    # it will slow this down substantially
    isolated_energies = {v: 0. for v in isolated}
    for term, bias in poly.items():

        isolated_components = term.intersection(isolated)

        if not isolated_components:
            continue

        en = bias  # energy contribution of the term
        for v in term:
            if v in isolated_energies:
                continue
            en *= samples[:, sampleset.variables.index(v)]

        for v in isolated_components:
            isolated_energies[v] += en

    # now put those energies into new_samples
    for col, v in enumerate(isolated):
        new_samples[:, col] = isolated_energies[v] < 0

    if poly.vartype is dimod.SPIN:
        new_samples = 2*new_samples - 1

    return np.concatenate((samples, new_samples), axis=1), list(variables) + isolated



# Copyright 2018 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import minorminer  # into this namespace
import dwave.embedding.chimera
import dwave.embedding.drawing
import dwave.embedding.pegasus
import dwave.embedding.exceptions

from dwave.embedding.diagnostic import diagnose_embedding, is_valid_embedding, verify_embedding

from dwave.embedding.chain_breaks import broken_chains
from dwave.embedding.chain_breaks import discard, majority_vote, weighted_random, MinimizeEnergy

from dwave.embedding.transforms import embed_bqm, embed_ising, embed_qubo, unembed_sampleset, EmbeddedStructure

from dwave.embedding.utils import target_to_source, chain_to_quadratic, chain_break_frequency



# Copyright 2019 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Composites that do batch operations for reverse annealing.
"""

import collections.abc as abc

import dimod
import numpy as np

__all__ = ['ReverseAdvanceComposite', 'ReverseBatchStatesComposite']


class ReverseAdvanceComposite(dimod.ComposedSampler):
    """
    Composite that reverse anneals an initial sample through a sequence of anneal
    schedules.

    If you do not specify an initial sample, a random sample is used for the first
    submission. By default, each subsequent submission selects the most-found
    lowest-energy sample as its initial state. If you set reinitialize_state to False,
    which makes each submission behave like a random walk, the subsequent submission
    selects the last returned sample as its initial state.

    Args:
       sampler (:obj:`dimod.Sampler`):
            A dimod sampler.

    Examples:
       This example runs 100 reverse anneals each for three schedules on a problem
       constructed by setting random :math:`\pm 1` values on a clique (complete
       graph) of 15 nodes, minor-embedded on a D-Wave system using the
       :class:`.DWaveCliqueSampler` sampler.

       >>> import dimod
       >>> from dwave.system import DWaveCliqueSampler, ReverseAdvanceComposite
       ...
       >>> sampler = DWaveCliqueSampler()     # doctest: +SKIP
       >>> sampler_reverse = ReverseAdvanceComposite(sampler)    # doctest: +SKIP
       >>> schedule = [[[0.0, 1.0], [t, 0.5], [20, 1.0]] for t in (5, 10, 15)]
       ...
       >>> bqm = dimod.generators.ran_r(1, 15)
       >>> init_samples = {i: -1 for i in range(15)}
       >>> sampleset = sampler_reverse.sample(bqm,
       ...                                    anneal_schedules=schedule,
       ...                                    initial_state=init_samples,
       ...                                    num_reads=100,
       ...                                    reinitialize_state=True)     # doctest: +SKIP

    """

    def __init__(self, child_sampler):
        self._children = [child_sampler]

    @property
    def children(self):
        return self._children

    @property
    def parameters(self):
        param = self.child.parameters.copy()
        param['schedules'] = []
        return param

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    def sample(self, bqm, anneal_schedules=None, **parameters):
        """Sample the binary quadratic model using reverse annealing along a given set 
        of anneal schedules.

        Args:
            bqm (:obj:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            anneal_schedules (list of lists, optional, default=[[[0, 1], [1, 0.35], [9, 0.35], [10, 1]]]): 
                Anneal schedules in order of submission. Each schedule is formatted 
                as a list of [time, s] pairs, in which time is in microseconds and s 
                is the normalized persistent current in the range [0,1].

            initial_state (dict, optional): 
                The state to reverse anneal from. If not provided, it will
                be randomly generated.

            **parameters:
                Parameters for the sampling method, specified by the child sampler.

        Returns:
            :obj:`dimod.SampleSet` that has initial_state and schedule_index fields.

        Examples:
           This example runs 100 reverse anneals each for three schedules on a problem
           constructed by setting random :math:`\pm 1` values on a clique (complete
           graph) of 15 nodes, minor-embedded on a D-Wave system using the
           :class:`.DWaveCliqueSampler` sampler.

           >>> import dimod
           >>> from dwave.system import DWaveCliqueSampler, ReverseAdvanceComposite
           ...
           >>> sampler = DWaveCliqueSampler()         # doctest: +SKIP
           >>> sampler_reverse = ReverseAdvanceComposite(sampler)    # doctest: +SKIP
           >>> schedule = [[[0.0, 1.0], [t, 0.5], [20, 1.0]] for t in (5, 10, 15)]
           ...
           >>> bqm = dimod.generators.ran_r(1, 15)
           >>> init_samples = {i: -1 for i in range(15)}
           >>> sampleset = sampler_reverse.sample(bqm,
           ...                                    anneal_schedules=schedule,
           ...                                    initial_state=init_samples,
           ...                                    num_reads=100,
           ...                                    reinitialize_state=True)  # doctest: +SKIP


        """
        child = self.child

        if anneal_schedules is None:
            anneal_schedules = [[[0, 1], [1, 0.35], [9, 0.35], [10, 1]]]    

        vartype_values = list(bqm.vartype.value)
        if 'initial_state' not in parameters:
            initial_state = dict(zip(list(bqm.variables), np.random.choice(vartype_values, len(bqm))))
        else:
            initial_state = parameters.pop('initial_state')

        if not isinstance(initial_state, abc.Mapping):
            raise TypeError("initial state provided must be a dict, but received {}".format(initial_state))

        if 'reinitialize_state' not in parameters:
            parameters['reinitialize_state'] = True

            if "answer_mode" in child.parameters:
                parameters['answer_mode'] = 'histogram'

        samplesets = None
        for schedule_idx, anneal_schedule in enumerate(anneal_schedules):
            sampleset = child.sample(bqm, anneal_schedule=anneal_schedule, initial_state=initial_state,
                                     **parameters)

            initial_state, _ = dimod.as_samples(initial_state)

            if 'initial_state' not in sampleset.record.dtype.names:
                init_state_vect = []

                if parameters['reinitialize_state']:
                    init_state_vect = [initial_state[0].copy() for i in range(len(sampleset.record.energy))]
                else:
                    # each sample is the next sample's initial state
                    init_state_vect.append(initial_state[0].copy())
                    for sample in sampleset.record.sample[:-1]:
                        init_state_vect.append(sample)

                sampleset = dimod.append_data_vectors(sampleset, initial_state=init_state_vect)
        
            if 'schedule_index' not in sampleset.record.dtype.names:
                schedule_index_vect = [schedule_idx] * len(sampleset.record.energy)
                sampleset = dimod.append_data_vectors(sampleset, schedule_index=schedule_index_vect)

            if samplesets is None:
                samplesets = sampleset
            else:
                samplesets = dimod.concatenate((samplesets, sampleset))

            if schedule_idx+1 == len(anneal_schedules):
                # no need to create the next initial state - last iteration
                break

            # prepare the initial state for the next iteration
            if parameters['reinitialize_state']:
                # if reinitialize is on, choose the lowest energy, most probable state for next iteration
                ground_state_energy = sampleset.first.energy
                lowest_energy_samples = sampleset.record[sampleset.record.energy == ground_state_energy]
                lowest_energy_samples.sort(order='num_occurrences')
                initial_state = dict(zip(sampleset.variables, lowest_energy_samples[-1].sample))
            else:
                # if not reinitialized, take the last state as the next initial state
                initial_state = dict(zip(sampleset.variables, sampleset.record.sample[-1]))

        samplesets.info['anneal_schedules'] = anneal_schedules
        return samplesets


class ReverseBatchStatesComposite(dimod.ComposedSampler, dimod.Initialized):
    """Composite that reverse anneals from multiple initial samples. Each submission is independent
    from one another.

    Args:
       sampler (:obj:`dimod.Sampler`):
            A dimod sampler.

    Examples:
       This example runs 100 reverse anneals each from two initial states on a problem
       constructed by setting random :math:`\pm 1` values on a clique (complete
       graph) of 15 nodes, minor-embedded on a D-Wave system using the
       :class:`.DWaveCliqueSampler` sampler.

       >>> import dimod
       >>> from dwave.system import DWaveCliqueSampler, ReverseBatchStatesComposite
       ...
       >>> sampler = DWaveCliqueSampler()      # doctest: +SKIP
       >>> sampler_reverse = ReverseBatchStatesComposite(sampler)    # doctest: +SKIP
       >>> schedule = [[0.0, 1.0], [10.0, 0.5], [20, 1.0]]
       ...
       >>> bqm = dimod.generators.ran_r(1, 15)
       >>> init_samples = [{i: -1 for i in range(15)}, {i: 1 for i in range(15)}]
       >>> sampleset = sampler_reverse.sample(bqm,
       ...                                    anneal_schedule=schedule,
       ...                                    initial_states=init_samples,
       ...                                    num_reads=100,
       ...                                    reinitialize_state=True)   # doctest: +SKIP



    """

    def __init__(self, child_sampler):
        self._children = [child_sampler]

    @property
    def children(self):
        return self._children

    @property
    def parameters(self):
        param = self.child.parameters.copy()
        param['initial_states'] = []
        return param

    @property
    def properties(self):
        return {'child_properties': self.child.properties.copy()}

    def sample(self, bqm, initial_states=None, initial_states_generator='random', num_reads=None, 
               seed=None, **parameters):
        """Sample the binary quadratic model using reverse annealing from multiple initial states.

        Args:
            bqm (:obj:`dimod.BinaryQuadraticModel`):
                Binary quadratic model to be sampled from.

            initial_states (samples-like, optional, default=None):
                One or more samples, each defining an initial state for all the problem variables. 
                If fewer than `num_reads` initial states are defined, additional values are 
                generated as specified by `initial_states_generator`. See :func:`dimod.as_samples` 
                for a description of "samples-like".

            initial_states_generator ({'none', 'tile', 'random'}, optional, default='random'):
                Defines the expansion of `initial_states` if fewer than
                `num_reads` are specified:

                * "none":
                    If the number of initial states specified is smaller than
                    `num_reads`, raises ValueError.

                * "tile":
                    Reuses the specified initial states if fewer than `num_reads`
                    or truncates if greater.

                * "random":
                    Expands the specified initial states with randomly generated
                    states if fewer than `num_reads` or truncates if greater.

            num_reads (int, optional, default=len(initial_states) or 1):
                Equivalent to number of desired initial states. If greater than the number of 
                provided initial states, additional states will be generated. If not provided, 
                it is selected to match the length of `initial_states`. If `initial_states` 
                is not provided, `num_reads` defaults to 1.

            seed (int (32-bit unsigned integer), optional):
                Seed to use for the PRNG. Specifying a particular seed with a
                constant set of parameters produces identical results. If not
                provided, a random seed is chosen.

            **parameters:
                Parameters for the sampling method, specified by the child sampler.

        Returns:
            :obj:`dimod.SampleSet` that has initial_state field.

        Examples:
           This example runs 100 reverse anneals each from two initial states on a problem
           constructed by setting random :math:`\pm 1` values on a clique (complete
           graph) of 15 nodes, minor-embedded on a D-Wave system using the
           :class:`.DWaveCliqueSampler` sampler.

           >>> import dimod
           >>> from dwave.system import DWaveCliqueSampler, ReverseBatchStatesComposite
           ...
           >>> sampler = DWaveCliqueSampler()       # doctest: +SKIP
           >>> sampler_reverse = ReverseBatchStatesComposite(sampler)   # doctest: +SKIP
           >>> schedule = [[0.0, 1.0], [10.0, 0.5], [20, 1.0]]
           ...
           >>> bqm = dimod.generators.ran_r(1, 15)
           >>> init_samples = [{i: -1 for i in range(15)}, {i: 1 for i in range(15)}]
           >>> sampleset = sampler_reverse.sample(bqm,
           ...                                    anneal_schedule=schedule,
           ...                                    initial_states=init_samples,
           ...                                    num_reads=100,
           ...                                    reinitialize_state=True)  # doctest: +SKIP


        """
        child = self.child
        
        parsed = self.parse_initial_states(bqm, 
                                           initial_states=initial_states,
                                           initial_states_generator=initial_states_generator,
                                           num_reads=num_reads,
                                           seed=seed)
        
        parsed_initial_states = np.ascontiguousarray(parsed.initial_states.record.sample)

        # there is gonna be way too much data generated - better to histogram them if possible
        if 'answer_mode' in child.parameters:
            parameters['answer_mode'] = 'histogram'

        # reduce number of network calls if possible by aggregating init states
        if 'num_reads' in child.parameters:
            aggreg = parsed.initial_states.aggregate()
            parsed_initial_states = np.ascontiguousarray(aggreg.record.sample)
            parameters['num_reads'] = aggreg.record.num_occurrences

        samplesets = None
        
        for initial_state in parsed_initial_states:
            sampleset = child.sample(bqm, initial_state=dict(zip(bqm.variables, initial_state)), **parameters)

            if 'initial_state' not in sampleset.record.dtype.names:
                init_state_vect = [initial_state.copy() for i in range(len(sampleset.record.energy))]
                sampleset = dimod.append_data_vectors(sampleset, initial_state=init_state_vect)

            if samplesets is None:
                samplesets = sampleset
            else:
                samplesets = dimod.concatenate((samplesets, sampleset))

        return samplesets
